# Inclui os membros

# Deprecated
from .c_v_rhq_s import CVRHQs  # noqa
from .c_v_rhq import CVRHQ  # noqa
from .c_v_rhv_s import CVRHVs  # noqa
from .c_v_rhv import CVRHV  # noqa
from .celetricas import Celetricas  # noqa
from .vdesviouh import Vdesviouh  # noqa
from .vturuh import Vturuh  # noqa
from .vertuh import Vertuh  # noqa
from .vagua import Vagua  # noqa
from .deletricas import Deletricas  # noqa
from .vevmin import Vevmin  # noqa
from .vevminm import Vevminm  # noqa
from .vevminsin import Vevminsin  # noqa
from .dfphauh import Dfphauh  # noqa
from .vghmin import Vghmin  # noqa
from .vghminm import Vghminm  # noqa
from .vghminsin import Vghminsin  # noqa
from .vghminuh import Vghminuh  # noqa
from .dlppdfmax import Dlppdfmax  # noqa
from .dlppdfmaxm import Dlppdfmaxm  # noqa
from .dlppdfmaxs import Dlppdfmaxs  # noqa
from .dlpptbmax import Dlpptbmax  # noqa
from .dlpptbmaxm import Dlpptbmaxm  # noqa
from .dlpptbmaxs import Dlpptbmaxs  # noqa
from .dtbmax import Dtbmax  # noqa
from .dtbmin import Dtbmin  # noqa
from .dvazmax import Dvazmax  # noqa
from .depminuh import Depminuh  # noqa
from .desvuh import Desvuh  # noqa
from .dposevap import Dposevap  # noqa
from .dnegevap import Dnegevap  # noqa
from .vbomb import Vbomb  # noqa

from .nwlistopdat import Nwlistopdat  # noqa
from .cviol_rhq_sin import CviolRhqsin  # noqa
from .cviol_rhq import CviolRhq  # noqa
from .cviol_rhv_sin import CviolRhvsin  # noqa
from .cviol_rhv import CviolRhv  # noqa
from .cbomb import Cbomb  # noqa
from .cbombsin import Cbombsin  # noqa
from .cdef import Cdef  # noqa
from .cdefsin import Cdefsin  # noqa
from .cviol_eletrica import CviolEletrica  # noqa
from .cviol_eletricasin import CviolEletricasin  # noqa
from .cmarg import Cmarg  # noqa
from .cmargmed import Cmargmed  # noqa
from .coper import Coper  # noqa
from .custo_futuro import CustoFuturo  # noqa
from .corteolm import Corteolm  # noqa
from .cterm import Cterm  # noqa
from .ctermsin import Ctermsin  # noqa
from .deficit import Def  # noqa
from .defsin import Defsin  # noqa
from .edesvc import Edesvc  # noqa
from .edesvcm import Edesvcm  # noqa
from .edesvcsin import Edesvcsin  # noqa
from .evapo import Evapo  # noqa
from .evapom import Evapom  # noqa
from .evaporsin import Evaporsin  # noqa
from .evert import Evert  # noqa
from .evertm import Evertm  # noqa
from .evertsin import Evertsin  # noqa
from .exces import Exces  # noqa
from .excessin import Excessin  # noqa
from .form_rhq import FormRHQ  # noqa
from .form_rhv import FormRHV  # noqa
from .fteolm import Fteolm  # noqa
from .fteolsin import Fteolsin  # noqa
from .eaf import Eaf  # noqa
from .eafb import Eafb  # noqa
from .eafbm import Eafbm  # noqa
from .eafbsin import Eafbsin  # noqa
from .eafm import Eafm  # noqa
from .earmf import Earmf  # noqa
from .earmfm import Earmfm  # noqa
from .earmfp import Earmfp  # noqa
from .earmfpm import Earmfpm  # noqa
from .earmfsin import Earmfsin  # noqa
from .earmfpsin import Earmfpsin  # noqa
from .geol import Geol  # noqa
from .geolm import Geolm  # noqa
from .geolsin import Geolsin  # noqa
from .ghiduh import Ghiduh  # noqa
from .ghmax_fpha import GhmaxFpha  # noqa
from .ghmax_fphc import GhmaxFphc  # noqa
from .ghmax import Ghmax  # noqa
from .ghmaxm import Ghmaxm  # noqa
from .ghmaxmr import Ghmaxmr  # noqa
from .ghmaxr import Ghmaxr  # noqa
from .ghmaxrsin import Ghmaxrsin  # noqa
from .ghmaxsin import Ghmaxsin  # noqa
from .ghidr import Ghidr  # noqa
from .ghidrm import Ghidrm  # noqa
from .ghidrsin import Ghidrsin  # noqa
from .ghtot import Ghtot  # noqa
from .ghtotm import Ghtotm  # noqa
from .ghtotsin import Ghtotsin  # noqa
from .gtert import Gtert  # noqa
from .gttot import Gttot  # noqa
from .gttotsin import Gttotsin  # noqa
from .hjus import Hjus  # noqa
from .hliq import Hliq  # noqa
from .hmont import Hmont  # noqa
from .intercambio import Intercambio  # noqa
from .invade import Invade  # noqa
from .invadem import Invadem  # noqa
from .mediassin import Mediassin  # noqa
from .mediasmerc import Mediasmerc  # noqa
from .mediasree import Mediasree  # noqa
from .mediasusih import Mediasusih  # noqa
from .mediasusit import Mediasusit  # noqa
from .mediasusie import Mediasusie  # noqa
from .mediasrep import Mediasrep  # noqa
from .mediasrhq import Mediasrhq  # noqa
from .mediasrhv import Mediasrhv  # noqa
from .mercl import Mercl  # noqa
from .merclsin import Merclsin  # noqa
from .mevmin import Mevmin  # noqa
from .mevminm import Mevminm  # noqa
from .mevminsin import Mevminsin  # noqa
from .perdf import Perdf  # noqa
from .perdfm import Perdfm  # noqa
from .perdfsin import Perdfsin  # noqa
from .pivarm import Pivarm  # noqa
from .pivarmincr import Pivarmincr  # noqa
from .qafluh import Qafluh  # noqa
from .qincruh import Qincruh  # noqa
from .rhslppdf import Rhslppdf  # noqa
from .rhslpptb import Rhslpptb  # noqa
from .valor_agua import ValorAgua  # noqa
from .varmpuh import Varmpuh  # noqa
from .varmuh import Varmuh  # noqa
from .qbomb import Qbomb  # noqa
from .vento import Vento  # noqa
from .qvertuh import Qvertuh  # noqa
from .verturb import Verturb  # noqa
from .verturbm import Verturbm  # noqa
from .verturbsin import Verturbsin  # noqa
from .qturuh import Qturuh  # noqa
from .qdesviouh import Qdesviouh  # noqa
from .vmort import Vmort  # noqa
from .vmortm import Vmortm  # noqa
from .vmortsin import Vmortsin  # noqa
from .vevapuh import Vevapuh  # noqa
from .viol_eletricasin import ViolEletricasin  # noqa
from .vretiradauh import Vretiradauh  # noqa
from .viol_fpha import ViolFpha  # noqa
from .viol_lpp_dfmax import ViolLppDfmax  # noqa
from .viol_lpp_dfmaxm import ViolLppDfmaxm  # noqa
from .viol_lpp_dfmaxsin import ViolLppDfmaxsin  # noqa
from .viol_lpp_tbmax import ViolLppTbmax  # noqa
from .viol_lpp_tbmaxm import ViolLppTbmaxm  # noqa
from .viol_lpp_tbmaxsin import ViolLppTbmaxsin  # noqa
from .viol_turbmax import ViolTurbmax  # noqa
from .viol_turbmin import ViolTurbmin  # noqa
from .viol_vazmax import ViolVazmax  # noqa
from .viol_vazmin import ViolVazmin  # noqa
from .viol_evmin import ViolEvmin  # noqa
from .viol_evminm import ViolEvminm  # noqa
from .viol_evminsin import ViolEvminsin  # noqa
from .viol_ghmin import ViolGhmin  # noqa
from .viol_ghminm import ViolGhminm  # noqa
from .viol_ghminsin import ViolGhminsin  # noqa
from .viol_ghminuh import ViolGhminuh  # noqa
from .viol_rhq import ViolRHQ  # noqa
from .viol_rhv import ViolRHV  # noqa
from .viol_eletrica import ViolEletrica  # noqa
from .viol_pos_evap import ViolPosEvap  # noqa
from .viol_neg_evap import ViolNegEvap  # noqa
from .viol_neg_vretiruh import ViolNegVretiruh  # noqa
from .viol_pos_vretiruh import ViolPosVretiruh  # noqa



from inewave.newave.modelos.pmo import BlocoVersaoModeloPMO
from inewave.newave.modelos.pmo import BlocoEafPastTendenciaHidrolPMO
from inewave.newave.modelos.pmo import BlocoEafPastCfugaMedioPMO
from inewave.newave.modelos.pmo import BlocoConvergenciaPMO
from inewave.newave.modelos.pmo import BlocoConfiguracoesExpansaoPMO
from inewave.newave.modelos.pmo import BlocoMARSPMO
from inewave.newave.modelos.pmo import BlocoRiscoDeficitENSPMO
from inewave.newave.modelos.pmo import BlocoCustoOperacaoPMO
from inewave.newave.modelos.pmo import BlocoCustoOperacaoTotalPMO
from inewave.newave.modelos.pmo import BlocoProdutibilidadesConfiguracaoPMO
from inewave.newave.modelos.pmo import BlocoEnergiaArmazenadaMaximaPMO
from inewave.newave.modelos.pmo import BlocoEnergiaArmazenadaInicialPMO
from inewave.newave.modelos.pmo import BlocoVolumeArmazenadoInicialPMO
from inewave.newave.modelos.pmo import BlocoPenalidadeViolacaoOutrosUsosPMO
from inewave.newave.modelos.pmo import BlocoPenalidadeViolacaoVazaoMinimaPMO
from inewave.newave.modelos.pmo import BlocoPenalidadeViolacaoCurvaSegurancaPMO
from inewave.newave.modelos.pmo import BlocoPenalidadeViolacaoFphaPMO
from inewave.newave.modelos.pmo import BlocoPenalidadeViolacaoEvaporacaoPMO
from inewave.newave.modelos.pmo import (
    BlocoPenalidadeViolacaoTurbinamentoMaximoPMO,
)
from inewave.newave.modelos.pmo import (
    BlocoPenalidadeViolacaoTurbinamentoMinimoPMO,
)
from inewave.newave.modelos.pmo import (
    BlocoGeracaoMinimaUsinasTermicasPMO,
    BlocoGeracaoMaximaUsinasTermicasPMO,
)

from cfinterface.files.blockfile import BlockFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Pmo(BlockFile):
    """
    Armazena os dados de entrada do NEWAVE referentes ao
    acompanhamento do programa.

    Esta classe lida com as informações de entrada fornecidas ao
    NEWAVE e reproduzidas no `pmo.dat`, bem como as saídas finais
    da execução: custos de operação, energias, déficit, etc.

    Em versões futuras, esta classe pode passar a ler os dados
    de execução intermediárias do programa.

    """

    T = TypeVar("T")

    BLOCKS = [
        BlocoVersaoModeloPMO,
        BlocoEafPastTendenciaHidrolPMO,
        BlocoEafPastCfugaMedioPMO,
        BlocoConvergenciaPMO,
        BlocoConfiguracoesExpansaoPMO,
        BlocoMARSPMO,
        BlocoRiscoDeficitENSPMO,
        BlocoCustoOperacaoPMO,
        BlocoCustoOperacaoTotalPMO,
        BlocoProdutibilidadesConfiguracaoPMO,
        BlocoEnergiaArmazenadaMaximaPMO,
        BlocoEnergiaArmazenadaInicialPMO,
        BlocoVolumeArmazenadoInicialPMO,
        BlocoPenalidadeViolacaoOutrosUsosPMO,
        BlocoPenalidadeViolacaoVazaoMinimaPMO,
        BlocoPenalidadeViolacaoCurvaSegurancaPMO,
        BlocoPenalidadeViolacaoFphaPMO,
        BlocoPenalidadeViolacaoEvaporacaoPMO,
        BlocoPenalidadeViolacaoTurbinamentoMaximoPMO,
        BlocoPenalidadeViolacaoTurbinamentoMinimoPMO,
        BlocoGeracaoMinimaUsinasTermicasPMO,
        BlocoGeracaoMaximaUsinasTermicasPMO,
    ]

    @property
    def versao_modelo(self) -> Optional[str]:
        """
        A versão do modelo que produziu o arquivo.

        :return: A string de versão do modelo.
        :rtype: str | None
        """
        b = self.data.get_blocks_of_type(BlocoVersaoModeloPMO)
        if isinstance(b, BlocoVersaoModeloPMO):
            return b.data
        elif isinstance(b, list):
            return b[0].data
        return None

    @property
    def eafpast_tendencia_hidrologica(self) -> Optional[pd.DataFrame]:
        """
        Energias afluentes passadas por REE para análise da tendência
        hidrológica, em relação à primeira configuração do sistema,
        em MWmes.

        - nome_ree (`str`)
        - mes (`int`)
        - valor (`float`)

        :return: A tendência hidrológica em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoEafPastTendenciaHidrolPMO)
        if isinstance(b, BlocoEafPastTendenciaHidrolPMO):
            return b.data
        return None

    @property
    def eafpast_cfuga_medio(self) -> Optional[pd.DataFrame]:
        """
        Energias afluentes passadas por REE considerando canal de
        fuga médio, em relação à primeira configuração do sistema,
        em MWmes.

        - nome_ree (`str`)
        - mes (`int`)
        - valor (`float`)

        :return: As energias afluentes passadas.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoEafPastCfugaMedioPMO)
        if isinstance(b, BlocoEafPastCfugaMedioPMO):
            return b.data
        return None

    @property
    def configuracoes_entrada_reservatorio(self) -> Optional[pd.DataFrame]:
        """
        Configurações do sistema em cada período devido a entrada
        de reservatórios e/ou potência de base.

        - data (`datetime`)
        - valor (`int`)

        :return: As configurações em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoConfiguracoesExpansaoPMO)
        if isinstance(b, list):
            return b[0].data
        return None

    @property
    def configuracoes_alteracao_potencia(self) -> Optional[pd.DataFrame]:
        """
        Configurações do sistema em cada período devido a alterações
        de potência.

        - data (`datetime`)
        - valor (`int`)

        :return: As configurações em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoConfiguracoesExpansaoPMO)
        if isinstance(b, list):
            return b[1].data
        return None

    @property
    def configuracoes_qualquer_modificacao(self) -> Optional[pd.DataFrame]:
        """
        Configurações do sistema em cada período devido a alterações
        de potência.

        - data (`datetime`)
        - valor (`int`)

        :return: As configurações em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoConfiguracoesExpansaoPMO)
        if isinstance(b, list):
            return b[2].data
        elif isinstance(b, BlocoConfiguracoesExpansaoPMO):
            return b.data
        return None

    @property
    def energia_armazenada_maxima(self) -> Optional[pd.DataFrame]:
        """
        Valores da energia armazenada máxima para cada REE do caso.

        - nome_ree (`str`)
        - configuracao (`int`)
        - valor_MWmes (`float`)

        :return: As energias em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoEnergiaArmazenadaMaximaPMO)
        if isinstance(b, BlocoEnergiaArmazenadaMaximaPMO):
            return b.data
        return None

    @property
    def energia_armazenada_inicial(self) -> Optional[pd.DataFrame]:
        """
        Valores da energia armazenada inicial para cada REE do caso.

        - nome_ree (`str`)
        - valor_MWmes (`float`)
        - valor_percentual (`float`)

        :return: As energias em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoEnergiaArmazenadaInicialPMO)
        if isinstance(b, BlocoEnergiaArmazenadaInicialPMO):
            return b.data
        return None

    @property
    def volume_armazenado_inicial(self) -> Optional[pd.DataFrame]:
        """
        Valores do volume armazenado inicial para cada UHE do caso.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - valor_hm3 (`float`)
        - valor_percentual (`float`)

        :return: As configurações em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoVolumeArmazenadoInicialPMO)
        if isinstance(b, BlocoVolumeArmazenadoInicialPMO):
            return b.data
        return None

    def retas_perdas_engolimento(self, estagio: int) -> Optional[pd.DataFrame]:
        """
        Retas ajustadas segundo o modelo MARS para corrigir a
        energia fio d'água com as perdas por engolimento máximo.

        - nome_ree (`str`)
        - reta (`int`)
        - coeficiente_angular (`float`)
        - coeficiente_linear (`float`)

        :return: As retas em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoMARSPMO)
        if isinstance(b, list):
            return b[estagio - 1].data
        return None

    @property
    def convergencia(self) -> Optional[pd.DataFrame]:
        """
        Tabela de convergência da execução do NEWAVE.

        - iteracao (`int`)
        - limite_inferior_zinf (`float`)
        - zinf (`float`)
        - limite_superior_zinf (`float`)
        - zsup (`float`)
        - delta_zinf (`float`)
        - zsup_iteracao (`float`)
        - tempo (`timedelta`)

        :return: As convergência em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoConvergenciaPMO)
        if isinstance(b, BlocoConvergenciaPMO):
            return b.data
        return None

    @property
    def risco_deficit_ens(self) -> Optional[pd.DataFrame]:
        """
        Tabela de riscos de déficit e enegia não suprida (ENS).

        - ano (`int`)
        - nome_submercado (`str`)
        - risco (`float`)
        - eens (`float`)

        :return: Os ricos em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoRiscoDeficitENSPMO)
        if isinstance(b, BlocoRiscoDeficitENSPMO):
            return b.data
        return None

    @property
    def custo_operacao_series_simuladas(self) -> Optional[pd.DataFrame]:
        """
        Tabela de custos de operação categorizados para as
        séries simuladas.

        - parcela (`str`)
        - valor_esperado (`float`)
        - desvio_padrao (`float`)
        - percentual (`float`)

        :return: Os custos em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoCustoOperacaoPMO)
        if isinstance(b, list):
            return b[0].data
        return None

    @property
    def valor_esperado_periodo_estudo(self) -> Optional[pd.DataFrame]:
        """
        Tabela de custos de operação esperados para o período
        de estudo.

        - parcela (`str`)
        - valor_esperado (`float`)
        - desvio_padrao (`float`)
        - percentual (`float`)

        :return: Os custos em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoCustoOperacaoPMO)
        if isinstance(b, list):
            return b[1].data
        return None

    @property
    def custo_operacao_referenciado_primeiro_mes(
        self,
    ) -> Optional[pd.DataFrame]:
        """
        Tabela de custos de operação esperados para o período
        de estudo, referenciados ao primeiro mês.

        - parcela (`str`)
        - valor_esperado (`float`)
        - desvio_padrao (`float`)
        - percentual (`float`)

        :return: Os custos em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoCustoOperacaoPMO)
        if isinstance(b, list):
            return b[2].data
        return None

    @property
    def custo_operacao_total(self) -> Optional[float]:
        """
        Custo de operacao total da SF.

        :return: O custo total.
        :rtype: float | None
        """
        b = self.data.get_blocks_of_type(BlocoCustoOperacaoTotalPMO)
        if isinstance(b, BlocoCustoOperacaoTotalPMO):
            return b.data[0]
        return None

    @property
    def desvio_custo_operacao_total(self) -> Optional[float]:
        """
        O desvio padrão do custo de operacao total da SF.

        :return: O desvio do custo total.
        :rtype: float | None
        """
        b = self.data.get_blocks_of_type(BlocoCustoOperacaoTotalPMO)
        if isinstance(b, BlocoCustoOperacaoTotalPMO):
            return b.data[1]
        return None

    @property
    def produtibilidades_equivalentes(self) -> Optional[pd.DataFrame]:
        """
        Tabela de produtibilidades calculadas para diversos fins do NEWAVE
        por usina e por configuração.

        - nome_usina (`str`)
        - configuracao (`int`)
        - produtibilidade_equivalente_volmin_volmax (`float`)
        - produtibilidade_equivalente_volmin_vol65 (`float`)
        - produtibilidade_altura_minima (`float`)
        - produtibilidade_altura_65 (`float`)
        - produtibilidade_altura_maxima (`float`)
        - produtibilidade_acumulada_calculo_earm (`float`)
        - produtibilidade_acumulada_calculo_earm_65 (`float`)
        - produtibilidade_acumulada_calculo_econ (`float`)
        - produtibilidade_acumulada_calculo_altura_minima (`float`)
        - produtibilidade_acumulada_calculo_altura_65 (`float`)
        - produtibilidade_acumulada_calculo_altura_maxima (`float`)
        - produtibilidade_acumulada_calculo_evaporacao_altura_minima (`float`)
        - produtibilidade_acumulada_calculo_evaporacao_altura_65 (`float`)
        - produtibilidade_acumulada_calculo_evaporacao_altura_maxima (`float`)


        :return: As produtibilidades em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoProdutibilidadesConfiguracaoPMO)
        if isinstance(b, BlocoProdutibilidadesConfiguracaoPMO):
            return b.data
        return None

    @property
    def penalidade_violacao_outros_usos(self) -> Optional[pd.DataFrame]:
        """
        Tabela de penalidades aplicadas à violação de restrições de
        outros usos da água.

        - ree (`str`)
        - data (`datetime`)
        - valor (`float`)

        :return: As penalidades em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoPenalidadeViolacaoOutrosUsosPMO)
        if isinstance(b, BlocoPenalidadeViolacaoOutrosUsosPMO):
            return b.data
        return None

    @property
    def penalidade_violacao_vazao_minima(self) -> Optional[pd.DataFrame]:
        """
        Tabela de penalidades aplicadas à violação de restrições de
        vazão mínima.

        - ree (`str`)
        - data (`datetime`)
        - patamar (`int`)
        - valor (`float`)

        :return: As penalidades em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoPenalidadeViolacaoVazaoMinimaPMO)
        if isinstance(b, BlocoPenalidadeViolacaoVazaoMinimaPMO):
            return b.data
        return None

    @property
    def penalidade_violacao_turbinamento_minimo(
        self,
    ) -> Optional[pd.DataFrame]:
        """
        Tabela de penalidades aplicadas à violação de restrições de
        turbinamento mínimo.

        - data (`datetime`)
        - valor (`float`)

        :return: As penalidades em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(
            BlocoPenalidadeViolacaoTurbinamentoMinimoPMO
        )
        if isinstance(b, BlocoPenalidadeViolacaoTurbinamentoMinimoPMO):
            return b.data
        return None

    @property
    def penalidade_violacao_turbinamento_maximo(
        self,
    ) -> Optional[pd.DataFrame]:
        """
        Tabela de penalidades aplicadas à violação de restrições de
        turbinamento máximo.

        - data (`datetime`)
        - valor (`float`)

        :return: As penalidades em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(
            BlocoPenalidadeViolacaoTurbinamentoMaximoPMO
        )
        if isinstance(b, BlocoPenalidadeViolacaoTurbinamentoMaximoPMO):
            return b.data
        return None

    @property
    def penalidade_violacao_curva(
        self,
    ) -> Optional[pd.DataFrame]:
        """
        Tabela de penalidades aplicadas à violação de restrições de
        curva de segurança.

        - ree (`str`)
        - data (`datetime`)
        - valor (`float`)

        :return: As penalidades em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(
            BlocoPenalidadeViolacaoCurvaSegurancaPMO
        )
        if isinstance(b, BlocoPenalidadeViolacaoCurvaSegurancaPMO):
            return b.data
        return None

    @property
    def penalidade_violacao_fpha(
        self,
    ) -> Optional[pd.DataFrame]:
        """
        Tabela de penalidades aplicadas à violação de restrições da
        FPHA.

        - ree (`str`)
        - data (`datetime`)
        - valor (`float`)

        :return: As penalidades em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoPenalidadeViolacaoFphaPMO)
        if isinstance(b, BlocoPenalidadeViolacaoFphaPMO):
            return b.data
        return None

    @property
    def penalidade_violacao_evaporacao(
        self,
    ) -> Optional[pd.DataFrame]:
        """
        Tabela de penalidades aplicadas à violação de restrições da
        evaporação.

        - ree (`str`)
        - data (`datetime`)
        - valor (`float`)

        :return: As penalidades em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoPenalidadeViolacaoEvaporacaoPMO)
        if isinstance(b, BlocoPenalidadeViolacaoEvaporacaoPMO):
            return b.data
        return None

    @property
    def geracao_minima_usinas_termicas(
        self,
    ) -> Optional[pd.DataFrame]:
        """
        Tabela de geração térmica mínima por usina térmica
        existente no caso.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - data (`datetime`)
        - valor_MWmed (`float`)

        :return: As gerações em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoGeracaoMinimaUsinasTermicasPMO)
        cols = ["codigo_usina", "nome_usina", "data", "valor_MWmed"]
        if isinstance(b, list):
            df = pd.concat([b.data for b in b], ignore_index=True)
            df = df.sort_values(by=["codigo_usina", "data"])
            return df[cols]
        elif isinstance(b, BlocoGeracaoMinimaUsinasTermicasPMO):
            df = b.data.sort_values(by=["codigo_usina", "data"])
            return df[cols]
        return None

    @property
    def geracao_maxima_usinas_termicas(
        self,
    ) -> Optional[pd.DataFrame]:
        """
        Tabela de geração térmica máxima por usina térmica
        existente no caso.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - data (`datetime`)
        - valor_MWmed (`float`)

        :return: As gerações em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoGeracaoMaximaUsinasTermicasPMO)
        cols = ["codigo_usina", "nome_usina", "data", "valor_MWmed"]
        if isinstance(b, list):
            df = pd.concat([b.data for b in b], ignore_index=True)
            df = df.sort_values(by=["codigo_usina", "data"])
            return df[cols]
        elif isinstance(b, BlocoGeracaoMaximaUsinasTermicasPMO):
            df = b.data.sort_values(by=["codigo_usina", "data"])
            return df[cols]
        return None



# Rotinas de testes associadas ao arquivo MEDIAS-REP.CSV do NWLISTOP
from inewave.nwlistop.mediasrep import Mediasrep


ARQ_TESTE = "tests/_arquivos/MEDIAS-REP.CSV"


def test_eq_mediasrep():
    leitor = Mediasrep.read(ARQ_TESTE)
    leitor2 = Mediasrep.read(ARQ_TESTE)
    assert leitor == leitor2
    assert leitor.valores.shape == (0, 2)



from inewave.newave.forward import Forward
import pandas as pd

ARQ_FORWARD = "./tests/mocks/arquivos/forward.dat"

TAMANHO_REGISTRO = 41264
NUMERO_CLASSES_TERMICAS_SUBMERCADOS = [34, 19, 43, 24]


def test_atributos_encontrados_forward():
    f = Forward.read(
        ARQ_FORWARD,
        tamanho_registro=TAMANHO_REGISTRO,
        numero_estagios=1,
        numero_forwards=2,
        numero_rees=12,
        numero_submercados=4,
        numero_total_submercados=5,
        numero_patamares_carga=3,
        numero_patamares_deficit=1,
        numero_agrupamentos_intercambio=6,
        numero_classes_termicas_submercados=NUMERO_CLASSES_TERMICAS_SUBMERCADOS,
        numero_usinas_hidreletricas=162,
        lag_maximo_usinas_gnl=2,
        numero_parques_eolicos_equivalentes=0,
        numero_estacoes_bombeamento=0,
        nomes_classes_termicas=[
            "ANGRA 1",
            "ANGRA 2",
            "BAIXADA FLU",
            "CUBATAO",
            "CUIABA G CC",
            "DAIA",
            "DO ATLAN_CSA",
            "DO ATLANTICO",
            "GNA I",
            "GNA P. ACU 3",
            "GOIANIA II",
            "IBIRITE",
            "JUIZ DE FORA",
            "LINHARES",
            "MARLIM AZUL",
            "N.PIRATINING",
            "NORTEFLU-1",
            "NORTEFLU-2",
            "NORTEFLU-3",
            "NORTEFLU-4",
            "ONCA PINTADA",
            "PALMEIRAS GO",
            "PIRAT.12 G",
            "R.SILVEIRA",
            "SEROPEDICA",
            "ST.CRUZ 34",
            "ST.CRUZ NOVA",
            "STA VITORIA",
            "T.NORTE 2",
            "TERMOMACAE",
            "TERMORIO",
            "TRES LAGOAS",
            "VIANA",
            "XAVANTES",
            "ARAUCARIA",
            "ARGENTINA 1",
            "ARGENTINA 1B",
            "ARGENTINA 2A",
            "ARGENTINA 2B",
            "ARGENTINA 2C",
            "ARGENTINA 2D",
            "CAMBARA",
            "CANDIOTA 3",
            "CANOAS",
            "CISFRAMA",
            "FIGUEIRA",
            "J.LACERDA A1",
            "J.LACERDA A2",
            "J.LACERDA B",
            "J.LACERDA C",
            "PAMPA SUL",
            "SAO SEPE",
            "URUGUAIANA",
            "ALTOS",
            "ARACATI",
            "BAHIA I",
            "BATURITE",
            "CAMACARI MII",
            "CAMACARI PI",
            "CAMPINA GDE",
            "CAMPO MAIOR",
            "CAUCAIA",
            "CRATO",
            "ENGUIA PECEM",
            "ERB CANDEIAS",
            "FORTALEZA",
            "GLOBAL I",
            "GLOBAL II",
            "IGUATU",
            "JUAZEIRO N",
            "MARACANAU I",
            "MARAMBAIA",
            "MURICY",
            "NAZARIA",
            "P. PECEM I",
            "P. PECEM II",
            "P. SERGIPE I",
            "PAU FERRO I",
            "PECEM II",
            "PERNAMBU_III",
            "PETROLINA",
            "POTIGUAR",
            "POTIGUAR III",
            "PROSPERI III",
            "PROSPERID II",
            "PROSPERIDADE",
            "SUAPE II",
            "SYKUE I",
            "TERMOBAHIA",
            "TERMOCABO",
            "TERMOCEARA",
            "TERMOMANAUS",
            "TERMONE",
            "TERMOPB",
            "TERMOPE",
            "VALE DO ACU",
            "APARECIDA",
            "C. ROCHA",
            "GERAMAR I",
            "GERAMAR II",
            "JARAQUI",
            "MANAUARA",
            "MARAN IV L22",
            "MARAN IV L7",
            "MARAN V L22",
            "MARAN V L7",
            "MARANHAO III",
            "MARANHAO IV",
            "MARANHAO V",
            "MAUA 3",
            "N.VEN 2 L22",
            "N.VEN 2 L7",
            "N.VENECIA 2",
            "NT BARCARENA",
            "O. CANOAS 1",
            "PARNAIBA IV",
            "PARNAIBA V",
            "PONTA NEGRA",
            "PORTO ITAQUI",
            "TAMBAQUI",
        ],
        nomes_usinas_hidreletricas=[
            "FUNIL-GRANDE",
            "BATALHA",
            "SERRA FACAO",
            "CAPIM BRANC1",
            "CAPIM BRANC2",
            "CORUMBA IV",
            "PIRAJU",
            "ITAIPU",
            "NILO PECANHA",
            "FONTES",
            "BAGUARI",
            "SAO DOMINGOS",
            "RETIRO BAIXO",
            "TRES MARIAS",
            "QUEIMADO",
            "JAURU",
            "GUAPORE",
            "CORUMBA III",
            "SINOP",
            "COLIDER",
            "TELES PIRES",
            "SAO MANOEL",
            "SLT VERDINHO",
            "OURINHOS",
            "SERRA MESA",
            "CANA BRAVA",
            "SAO SALVADOR",
            "PEIXE ANGIC",
            "LAJEADO",
            "SALTO",
            "RONDON II",
            "PONTE PEDRA",
            "JIRAU",
            "STO ANTONIO",
            "ESPORA",
            "ITIQUIRA I",
            "ITIQUIRA II",
            "DARDANELOS",
            "CACU",
            "B. COQUEIROS",
            "FOZ R. CLARO",
            "FICT.MAUA",
            "GUARAPIRANGA",
            "BILLINGS",
            "HENRY BORDEN",
            "JAGUARI",
            "PARAIBUNA",
            "SANTA BRANCA",
            "FUNIL",
            "LAJES",
            "PICADA",
            "SOBRAGI",
            "SIMPLICIO",
            "ILHA POMBOS",
            "P. PASSOS",
            "SALTO GRANDE",
            "P. ESTRELA",
            "CANDONGA",
            "AIMORES",
            "MASCARENHAS",
            "GUILMAN-AMOR",
            "SA CARVALHO",
            "ROSAL",
            "SAMUEL",
            "IRAPE",
            "STA CLARA MG",
            "CAMARGOS",
            "ITUTINGA",
            "FURNAS",
            "M. DE MORAES",
            "ESTREITO",
            "JAGUARA",
            "IGARAPAVA",
            "VOLTA GRANDE",
            "P. COLOMBIA",
            "CACONDE",
            "E. DA CUNHA",
            "A.S.OLIVEIRA",
            "MARIMBONDO",
            "A. VERMELHA",
            "I. SOLTEIRA",
            "EMBORCACAO",
            "NOVA PONTE",
            "MIRANDA",
            "CORUMBA I",
            "ITUMBIARA",
            "CACH.DOURADA",
            "SAO SIMAO",
            "BARRA BONITA",
            "A.S. LIMA",
            "IBITINGA",
            "PROMISSAO",
            "NAVANHANDAVA",
            "TRES IRMAOS",
            "JUPIA",
            "P. PRIMAVERA",
            "MANSO",
            "A.A. LAYDNER",
            "CHAVANTES",
            "L.N. GARCEZ",
            "CANOAS II",
            "CANOAS I",
            "CAPIVARA",
            "TAQUARUCU",
            "ROSANA",
            "MAUA",
            "STA CLARA PR",
            "FUNDAO",
            "G.B. MUNHOZ",
            "BAIXO IGUACU",
            "BARRA GRANDE",
            "GARIBALDI",
            "CAMPOS NOVOS",
            "MACHADINHO",
            "ITA",
            "PASSO FUNDO",
            "MONJOLINHO",
            "QUEBRA QUEIX",
            "SAO JOSE",
            "PASSO S JOAO",
            "FOZ CHAPECO",
            "CASTRO ALVES",
            "MONTE CLARO",
            "14 DE JULHO",
            "ERNESTINA",
            "PASSO REAL",
            "JACUI",
            "ITAUBA",
            "D. FRANCISCA",
            "G.P. SOUZA",
            "SALTO PILAO",
            "JORDAO",
            "SEGREDO",
            "SLT.SANTIAGO",
            "SALTO OSORIO",
            "SALTO CAXIAS",
            "SOBRADINHO",
            "ITAPARICA",
            "COMP PAF-MOX",
            "XINGO",
            "FICT.QUEIMAD",
            "FICT.TRES MA",
            "FICT.RETIRO",
            "ITAPEBI",
            "FICT.IRAPE",
            "P. CAVALO",
            "B. ESPERANCA",
            "CACH.CALDEIR",
            "ESTREITO TOC",
            "TUCURUI",
            "BALBINA",
            "COARACY NUNE",
            "FERREIRA GOM",
            "STO ANT JARI",
            "FICT.SERRA M",
            "FICT.CANA BR",
            "FICT.LAJEADO",
            "FICT.PEIXE A",
            "FICT.SAO SAL",
            "PIMENTAL",
            "CURUA-UNA",
            "BELO MONTE",
        ],
    )
    assert isinstance(f.volume_armazenado_absoluto_final, pd.DataFrame)


def test_eq_forward():
    f1 = Forward.read(
        ARQ_FORWARD,
        tamanho_registro=TAMANHO_REGISTRO,
        numero_estagios=1,
        numero_forwards=2,
        numero_rees=12,
        numero_submercados=4,
        numero_total_submercados=5,
        numero_patamares_carga=3,
        numero_patamares_deficit=1,
        numero_agrupamentos_intercambio=6,
        numero_classes_termicas_submercados=NUMERO_CLASSES_TERMICAS_SUBMERCADOS,
        numero_usinas_hidreletricas=162,
        lag_maximo_usinas_gnl=2,
        numero_parques_eolicos_equivalentes=0,
        numero_estacoes_bombeamento=0,
        nomes_classes_termicas=[
            "ANGRA 1",
            "ANGRA 2",
            "BAIXADA FLU",
            "CUBATAO",
            "CUIABA G CC",
            "DAIA",
            "DO ATLAN_CSA",
            "DO ATLANTICO",
            "GNA I",
            "GNA P. ACU 3",
            "GOIANIA II",
            "IBIRITE",
            "JUIZ DE FORA",
            "LINHARES",
            "MARLIM AZUL",
            "N.PIRATINING",
            "NORTEFLU-1",
            "NORTEFLU-2",
            "NORTEFLU-3",
            "NORTEFLU-4",
            "ONCA PINTADA",
            "PALMEIRAS GO",
            "PIRAT.12 G",
            "R.SILVEIRA",
            "SEROPEDICA",
            "ST.CRUZ 34",
            "ST.CRUZ NOVA",
            "STA VITORIA",
            "T.NORTE 2",
            "TERMOMACAE",
            "TERMORIO",
            "TRES LAGOAS",
            "VIANA",
            "XAVANTES",
            "ARAUCARIA",
            "ARGENTINA 1",
            "ARGENTINA 1B",
            "ARGENTINA 2A",
            "ARGENTINA 2B",
            "ARGENTINA 2C",
            "ARGENTINA 2D",
            "CAMBARA",
            "CANDIOTA 3",
            "CANOAS",
            "CISFRAMA",
            "FIGUEIRA",
            "J.LACERDA A1",
            "J.LACERDA A2",
            "J.LACERDA B",
            "J.LACERDA C",
            "PAMPA SUL",
            "SAO SEPE",
            "URUGUAIANA",
            "ALTOS",
            "ARACATI",
            "BAHIA I",
            "BATURITE",
            "CAMACARI MII",
            "CAMACARI PI",
            "CAMPINA GDE",
            "CAMPO MAIOR",
            "CAUCAIA",
            "CRATO",
            "ENGUIA PECEM",
            "ERB CANDEIAS",
            "FORTALEZA",
            "GLOBAL I",
            "GLOBAL II",
            "IGUATU",
            "JUAZEIRO N",
            "MARACANAU I",
            "MARAMBAIA",
            "MURICY",
            "NAZARIA",
            "P. PECEM I",
            "P. PECEM II",
            "P. SERGIPE I",
            "PAU FERRO I",
            "PECEM II",
            "PERNAMBU_III",
            "PETROLINA",
            "POTIGUAR",
            "POTIGUAR III",
            "PROSPERI III",
            "PROSPERID II",
            "PROSPERIDADE",
            "SUAPE II",
            "SYKUE I",
            "TERMOBAHIA",
            "TERMOCABO",
            "TERMOCEARA",
            "TERMOMANAUS",
            "TERMONE",
            "TERMOPB",
            "TERMOPE",
            "VALE DO ACU",
            "APARECIDA",
            "C. ROCHA",
            "GERAMAR I",
            "GERAMAR II",
            "JARAQUI",
            "MANAUARA",
            "MARAN IV L22",
            "MARAN IV L7",
            "MARAN V L22",
            "MARAN V L7",
            "MARANHAO III",
            "MARANHAO IV",
            "MARANHAO V",
            "MAUA 3",
            "N.VEN 2 L22",
            "N.VEN 2 L7",
            "N.VENECIA 2",
            "NT BARCARENA",
            "O. CANOAS 1",
            "PARNAIBA IV",
            "PARNAIBA V",
            "PONTA NEGRA",
            "PORTO ITAQUI",
            "TAMBAQUI",
        ],
        nomes_usinas_hidreletricas=[
            "FUNIL-GRANDE",
            "BATALHA",
            "SERRA FACAO",
            "CAPIM BRANC1",
            "CAPIM BRANC2",
            "CORUMBA IV",
            "PIRAJU",
            "ITAIPU",
            "NILO PECANHA",
            "FONTES",
            "BAGUARI",
            "SAO DOMINGOS",
            "RETIRO BAIXO",
            "TRES MARIAS",
            "QUEIMADO",
            "JAURU",
            "GUAPORE",
            "CORUMBA III",
            "SINOP",
            "COLIDER",
            "TELES PIRES",
            "SAO MANOEL",
            "SLT VERDINHO",
            "OURINHOS",
            "SERRA MESA",
            "CANA BRAVA",
            "SAO SALVADOR",
            "PEIXE ANGIC",
            "LAJEADO",
            "SALTO",
            "RONDON II",
            "PONTE PEDRA",
            "JIRAU",
            "STO ANTONIO",
            "ESPORA",
            "ITIQUIRA I",
            "ITIQUIRA II",
            "DARDANELOS",
            "CACU",
            "B. COQUEIROS",
            "FOZ R. CLARO",
            "FICT.MAUA",
            "GUARAPIRANGA",
            "BILLINGS",
            "HENRY BORDEN",
            "JAGUARI",
            "PARAIBUNA",
            "SANTA BRANCA",
            "FUNIL",
            "LAJES",
            "PICADA",
            "SOBRAGI",
            "SIMPLICIO",
            "ILHA POMBOS",
            "P. PASSOS",
            "SALTO GRANDE",
            "P. ESTRELA",
            "CANDONGA",
            "AIMORES",
            "MASCARENHAS",
            "GUILMAN-AMOR",
            "SA CARVALHO",
            "ROSAL",
            "SAMUEL",
            "IRAPE",
            "STA CLARA MG",
            "CAMARGOS",
            "ITUTINGA",
            "FURNAS",
            "M. DE MORAES",
            "ESTREITO",
            "JAGUARA",
            "IGARAPAVA",
            "VOLTA GRANDE",
            "P. COLOMBIA",
            "CACONDE",
            "E. DA CUNHA",
            "A.S.OLIVEIRA",
            "MARIMBONDO",
            "A. VERMELHA",
            "I. SOLTEIRA",
            "EMBORCACAO",
            "NOVA PONTE",
            "MIRANDA",
            "CORUMBA I",
            "ITUMBIARA",
            "CACH.DOURADA",
            "SAO SIMAO",
            "BARRA BONITA",
            "A.S. LIMA",
            "IBITINGA",
            "PROMISSAO",
            "NAVANHANDAVA",
            "TRES IRMAOS",
            "JUPIA",
            "P. PRIMAVERA",
            "MANSO",
            "A.A. LAYDNER",
            "CHAVANTES",
            "L.N. GARCEZ",
            "CANOAS II",
            "CANOAS I",
            "CAPIVARA",
            "TAQUARUCU",
            "ROSANA",
            "MAUA",
            "STA CLARA PR",
            "FUNDAO",
            "G.B. MUNHOZ",
            "BAIXO IGUACU",
            "BARRA GRANDE",
            "GARIBALDI",
            "CAMPOS NOVOS",
            "MACHADINHO",
            "ITA",
            "PASSO FUNDO",
            "MONJOLINHO",
            "QUEBRA QUEIX",
            "SAO JOSE",
            "PASSO S JOAO",
            "FOZ CHAPECO",
            "CASTRO ALVES",
            "MONTE CLARO",
            "14 DE JULHO",
            "ERNESTINA",
            "PASSO REAL",
            "JACUI",
            "ITAUBA",
            "D. FRANCISCA",
            "G.P. SOUZA",
            "SALTO PILAO",
            "JORDAO",
            "SEGREDO",
            "SLT.SANTIAGO",
            "SALTO OSORIO",
            "SALTO CAXIAS",
            "SOBRADINHO",
            "ITAPARICA",
            "COMP PAF-MOX",
            "XINGO",
            "FICT.QUEIMAD",
            "FICT.TRES MA",
            "FICT.RETIRO",
            "ITAPEBI",
            "FICT.IRAPE",
            "P. CAVALO",
            "B. ESPERANCA",
            "CACH.CALDEIR",
            "ESTREITO TOC",
            "TUCURUI",
            "BALBINA",
            "COARACY NUNE",
            "FERREIRA GOM",
            "STO ANT JARI",
            "FICT.SERRA M",
            "FICT.CANA BR",
            "FICT.LAJEADO",
            "FICT.PEIXE A",
            "FICT.SAO SAL",
            "PIMENTAL",
            "CURUA-UNA",
            "BELO MONTE",
        ],
    )
    f2 = Forward.read(
        ARQ_FORWARD,
        tamanho_registro=TAMANHO_REGISTRO,
        numero_estagios=1,
        numero_forwards=2,
        numero_rees=12,
        numero_submercados=4,
        numero_total_submercados=5,
        numero_patamares_carga=3,
        numero_patamares_deficit=1,
        numero_agrupamentos_intercambio=6,
        numero_classes_termicas_submercados=NUMERO_CLASSES_TERMICAS_SUBMERCADOS,
        numero_usinas_hidreletricas=162,
        lag_maximo_usinas_gnl=2,
        numero_parques_eolicos_equivalentes=0,
        numero_estacoes_bombeamento=0,
        nomes_classes_termicas=[
            "ANGRA 1",
            "ANGRA 2",
            "BAIXADA FLU",
            "CUBATAO",
            "CUIABA G CC",
            "DAIA",
            "DO ATLAN_CSA",
            "DO ATLANTICO",
            "GNA I",
            "GNA P. ACU 3",
            "GOIANIA II",
            "IBIRITE",
            "JUIZ DE FORA",
            "LINHARES",
            "MARLIM AZUL",
            "N.PIRATINING",
            "NORTEFLU-1",
            "NORTEFLU-2",
            "NORTEFLU-3",
            "NORTEFLU-4",
            "ONCA PINTADA",
            "PALMEIRAS GO",
            "PIRAT.12 G",
            "R.SILVEIRA",
            "SEROPEDICA",
            "ST.CRUZ 34",
            "ST.CRUZ NOVA",
            "STA VITORIA",
            "T.NORTE 2",
            "TERMOMACAE",
            "TERMORIO",
            "TRES LAGOAS",
            "VIANA",
            "XAVANTES",
            "ARAUCARIA",
            "ARGENTINA 1",
            "ARGENTINA 1B",
            "ARGENTINA 2A",
            "ARGENTINA 2B",
            "ARGENTINA 2C",
            "ARGENTINA 2D",
            "CAMBARA",
            "CANDIOTA 3",
            "CANOAS",
            "CISFRAMA",
            "FIGUEIRA",
            "J.LACERDA A1",
            "J.LACERDA A2",
            "J.LACERDA B",
            "J.LACERDA C",
            "PAMPA SUL",
            "SAO SEPE",
            "URUGUAIANA",
            "ALTOS",
            "ARACATI",
            "BAHIA I",
            "BATURITE",
            "CAMACARI MII",
            "CAMACARI PI",
            "CAMPINA GDE",
            "CAMPO MAIOR",
            "CAUCAIA",
            "CRATO",
            "ENGUIA PECEM",
            "ERB CANDEIAS",
            "FORTALEZA",
            "GLOBAL I",
            "GLOBAL II",
            "IGUATU",
            "JUAZEIRO N",
            "MARACANAU I",
            "MARAMBAIA",
            "MURICY",
            "NAZARIA",
            "P. PECEM I",
            "P. PECEM II",
            "P. SERGIPE I",
            "PAU FERRO I",
            "PECEM II",
            "PERNAMBU_III",
            "PETROLINA",
            "POTIGUAR",
            "POTIGUAR III",
            "PROSPERI III",
            "PROSPERID II",
            "PROSPERIDADE",
            "SUAPE II",
            "SYKUE I",
            "TERMOBAHIA",
            "TERMOCABO",
            "TERMOCEARA",
            "TERMOMANAUS",
            "TERMONE",
            "TERMOPB",
            "TERMOPE",
            "VALE DO ACU",
            "APARECIDA",
            "C. ROCHA",
            "GERAMAR I",
            "GERAMAR II",
            "JARAQUI",
            "MANAUARA",
            "MARAN IV L22",
            "MARAN IV L7",
            "MARAN V L22",
            "MARAN V L7",
            "MARANHAO III",
            "MARANHAO IV",
            "MARANHAO V",
            "MAUA 3",
            "N.VEN 2 L22",
            "N.VEN 2 L7",
            "N.VENECIA 2",
            "NT BARCARENA",
            "O. CANOAS 1",
            "PARNAIBA IV",
            "PARNAIBA V",
            "PONTA NEGRA",
            "PORTO ITAQUI",
            "TAMBAQUI",
        ],
        nomes_usinas_hidreletricas=[
            "FUNIL-GRANDE",
            "BATALHA",
            "SERRA FACAO",
            "CAPIM BRANC1",
            "CAPIM BRANC2",
            "CORUMBA IV",
            "PIRAJU",
            "ITAIPU",
            "NILO PECANHA",
            "FONTES",
            "BAGUARI",
            "SAO DOMINGOS",
            "RETIRO BAIXO",
            "TRES MARIAS",
            "QUEIMADO",
            "JAURU",
            "GUAPORE",
            "CORUMBA III",
            "SINOP",
            "COLIDER",
            "TELES PIRES",
            "SAO MANOEL",
            "SLT VERDINHO",
            "OURINHOS",
            "SERRA MESA",
            "CANA BRAVA",
            "SAO SALVADOR",
            "PEIXE ANGIC",
            "LAJEADO",
            "SALTO",
            "RONDON II",
            "PONTE PEDRA",
            "JIRAU",
            "STO ANTONIO",
            "ESPORA",
            "ITIQUIRA I",
            "ITIQUIRA II",
            "DARDANELOS",
            "CACU",
            "B. COQUEIROS",
            "FOZ R. CLARO",
            "FICT.MAUA",
            "GUARAPIRANGA",
            "BILLINGS",
            "HENRY BORDEN",
            "JAGUARI",
            "PARAIBUNA",
            "SANTA BRANCA",
            "FUNIL",
            "LAJES",
            "PICADA",
            "SOBRAGI",
            "SIMPLICIO",
            "ILHA POMBOS",
            "P. PASSOS",
            "SALTO GRANDE",
            "P. ESTRELA",
            "CANDONGA",
            "AIMORES",
            "MASCARENHAS",
            "GUILMAN-AMOR",
            "SA CARVALHO",
            "ROSAL",
            "SAMUEL",
            "IRAPE",
            "STA CLARA MG",
            "CAMARGOS",
            "ITUTINGA",
            "FURNAS",
            "M. DE MORAES",
            "ESTREITO",
            "JAGUARA",
            "IGARAPAVA",
            "VOLTA GRANDE",
            "P. COLOMBIA",
            "CACONDE",
            "E. DA CUNHA",
            "A.S.OLIVEIRA",
            "MARIMBONDO",
            "A. VERMELHA",
            "I. SOLTEIRA",
            "EMBORCACAO",
            "NOVA PONTE",
            "MIRANDA",
            "CORUMBA I",
            "ITUMBIARA",
            "CACH.DOURADA",
            "SAO SIMAO",
            "BARRA BONITA",
            "A.S. LIMA",
            "IBITINGA",
            "PROMISSAO",
            "NAVANHANDAVA",
            "TRES IRMAOS",
            "JUPIA",
            "P. PRIMAVERA",
            "MANSO",
            "A.A. LAYDNER",
            "CHAVANTES",
            "L.N. GARCEZ",
            "CANOAS II",
            "CANOAS I",
            "CAPIVARA",
            "TAQUARUCU",
            "ROSANA",
            "MAUA",
            "STA CLARA PR",
            "FUNDAO",
            "G.B. MUNHOZ",
            "BAIXO IGUACU",
            "BARRA GRANDE",
            "GARIBALDI",
            "CAMPOS NOVOS",
            "MACHADINHO",
            "ITA",
            "PASSO FUNDO",
            "MONJOLINHO",
            "QUEBRA QUEIX",
            "SAO JOSE",
            "PASSO S JOAO",
            "FOZ CHAPECO",
            "CASTRO ALVES",
            "MONTE CLARO",
            "14 DE JULHO",
            "ERNESTINA",
            "PASSO REAL",
            "JACUI",
            "ITAUBA",
            "D. FRANCISCA",
            "G.P. SOUZA",
            "SALTO PILAO",
            "JORDAO",
            "SEGREDO",
            "SLT.SANTIAGO",
            "SALTO OSORIO",
            "SALTO CAXIAS",
            "SOBRADINHO",
            "ITAPARICA",
            "COMP PAF-MOX",
            "XINGO",
            "FICT.QUEIMAD",
            "FICT.TRES MA",
            "FICT.RETIRO",
            "ITAPEBI",
            "FICT.IRAPE",
            "P. CAVALO",
            "B. ESPERANCA",
            "CACH.CALDEIR",
            "ESTREITO TOC",
            "TUCURUI",
            "BALBINA",
            "COARACY NUNE",
            "FERREIRA GOM",
            "STO ANT JARI",
            "FICT.SERRA M",
            "FICT.CANA BR",
            "FICT.LAJEADO",
            "FICT.PEIXE A",
            "FICT.SAO SAL",
            "PIMENTAL",
            "CURUA-UNA",
            "BELO MONTE",
        ],
    )
    assert f1 == f2


def test_atributos_forward():
    f = Forward.read(
        ARQ_FORWARD,
        tamanho_registro=TAMANHO_REGISTRO,
        numero_estagios=1,
        numero_forwards=2,
        numero_rees=12,
        numero_submercados=4,
        numero_total_submercados=5,
        numero_patamares_carga=3,
        numero_patamares_deficit=1,
        numero_agrupamentos_intercambio=6,
        numero_classes_termicas_submercados=NUMERO_CLASSES_TERMICAS_SUBMERCADOS,
        numero_usinas_hidreletricas=162,
        lag_maximo_usinas_gnl=2,
        numero_parques_eolicos_equivalentes=0,
        numero_estacoes_bombeamento=0,
        nomes_classes_termicas=[
            "ANGRA 1",
            "ANGRA 2",
            "BAIXADA FLU",
            "CUBATAO",
            "CUIABA G CC",
            "DAIA",
            "DO ATLAN_CSA",
            "DO ATLANTICO",
            "GNA I",
            "GNA P. ACU 3",
            "GOIANIA II",
            "IBIRITE",
            "JUIZ DE FORA",
            "LINHARES",
            "MARLIM AZUL",
            "N.PIRATINING",
            "NORTEFLU-1",
            "NORTEFLU-2",
            "NORTEFLU-3",
            "NORTEFLU-4",
            "ONCA PINTADA",
            "PALMEIRAS GO",
            "PIRAT.12 G",
            "R.SILVEIRA",
            "SEROPEDICA",
            "ST.CRUZ 34",
            "ST.CRUZ NOVA",
            "STA VITORIA",
            "T.NORTE 2",
            "TERMOMACAE",
            "TERMORIO",
            "TRES LAGOAS",
            "VIANA",
            "XAVANTES",
            "ARAUCARIA",
            "ARGENTINA 1",
            "ARGENTINA 1B",
            "ARGENTINA 2A",
            "ARGENTINA 2B",
            "ARGENTINA 2C",
            "ARGENTINA 2D",
            "CAMBARA",
            "CANDIOTA 3",
            "CANOAS",
            "CISFRAMA",
            "FIGUEIRA",
            "J.LACERDA A1",
            "J.LACERDA A2",
            "J.LACERDA B",
            "J.LACERDA C",
            "PAMPA SUL",
            "SAO SEPE",
            "URUGUAIANA",
            "ALTOS",
            "ARACATI",
            "BAHIA I",
            "BATURITE",
            "CAMACARI MII",
            "CAMACARI PI",
            "CAMPINA GDE",
            "CAMPO MAIOR",
            "CAUCAIA",
            "CRATO",
            "ENGUIA PECEM",
            "ERB CANDEIAS",
            "FORTALEZA",
            "GLOBAL I",
            "GLOBAL II",
            "IGUATU",
            "JUAZEIRO N",
            "MARACANAU I",
            "MARAMBAIA",
            "MURICY",
            "NAZARIA",
            "P. PECEM I",
            "P. PECEM II",
            "P. SERGIPE I",
            "PAU FERRO I",
            "PECEM II",
            "PERNAMBU_III",
            "PETROLINA",
            "POTIGUAR",
            "POTIGUAR III",
            "PROSPERI III",
            "PROSPERID II",
            "PROSPERIDADE",
            "SUAPE II",
            "SYKUE I",
            "TERMOBAHIA",
            "TERMOCABO",
            "TERMOCEARA",
            "TERMOMANAUS",
            "TERMONE",
            "TERMOPB",
            "TERMOPE",
            "VALE DO ACU",
            "APARECIDA",
            "C. ROCHA",
            "GERAMAR I",
            "GERAMAR II",
            "JARAQUI",
            "MANAUARA",
            "MARAN IV L22",
            "MARAN IV L7",
            "MARAN V L22",
            "MARAN V L7",
            "MARANHAO III",
            "MARANHAO IV",
            "MARANHAO V",
            "MAUA 3",
            "N.VEN 2 L22",
            "N.VEN 2 L7",
            "N.VENECIA 2",
            "NT BARCARENA",
            "O. CANOAS 1",
            "PARNAIBA IV",
            "PARNAIBA V",
            "PONTA NEGRA",
            "PORTO ITAQUI",
            "TAMBAQUI",
        ],
        nomes_usinas_hidreletricas=[
            "FUNIL-GRANDE",
            "BATALHA",
            "SERRA FACAO",
            "CAPIM BRANC1",
            "CAPIM BRANC2",
            "CORUMBA IV",
            "PIRAJU",
            "ITAIPU",
            "NILO PECANHA",
            "FONTES",
            "BAGUARI",
            "SAO DOMINGOS",
            "RETIRO BAIXO",
            "TRES MARIAS",
            "QUEIMADO",
            "JAURU",
            "GUAPORE",
            "CORUMBA III",
            "SINOP",
            "COLIDER",
            "TELES PIRES",
            "SAO MANOEL",
            "SLT VERDINHO",
            "OURINHOS",
            "SERRA MESA",
            "CANA BRAVA",
            "SAO SALVADOR",
            "PEIXE ANGIC",
            "LAJEADO",
            "SALTO",
            "RONDON II",
            "PONTE PEDRA",
            "JIRAU",
            "STO ANTONIO",
            "ESPORA",
            "ITIQUIRA I",
            "ITIQUIRA II",
            "DARDANELOS",
            "CACU",
            "B. COQUEIROS",
            "FOZ R. CLARO",
            "FICT.MAUA",
            "GUARAPIRANGA",
            "BILLINGS",
            "HENRY BORDEN",
            "JAGUARI",
            "PARAIBUNA",
            "SANTA BRANCA",
            "FUNIL",
            "LAJES",
            "PICADA",
            "SOBRAGI",
            "SIMPLICIO",
            "ILHA POMBOS",
            "P. PASSOS",
            "SALTO GRANDE",
            "P. ESTRELA",
            "CANDONGA",
            "AIMORES",
            "MASCARENHAS",
            "GUILMAN-AMOR",
            "SA CARVALHO",
            "ROSAL",
            "SAMUEL",
            "IRAPE",
            "STA CLARA MG",
            "CAMARGOS",
            "ITUTINGA",
            "FURNAS",
            "M. DE MORAES",
            "ESTREITO",
            "JAGUARA",
            "IGARAPAVA",
            "VOLTA GRANDE",
            "P. COLOMBIA",
            "CACONDE",
            "E. DA CUNHA",
            "A.S.OLIVEIRA",
            "MARIMBONDO",
            "A. VERMELHA",
            "I. SOLTEIRA",
            "EMBORCACAO",
            "NOVA PONTE",
            "MIRANDA",
            "CORUMBA I",
            "ITUMBIARA",
            "CACH.DOURADA",
            "SAO SIMAO",
            "BARRA BONITA",
            "A.S. LIMA",
            "IBITINGA",
            "PROMISSAO",
            "NAVANHANDAVA",
            "TRES IRMAOS",
            "JUPIA",
            "P. PRIMAVERA",
            "MANSO",
            "A.A. LAYDNER",
            "CHAVANTES",
            "L.N. GARCEZ",
            "CANOAS II",
            "CANOAS I",
            "CAPIVARA",
            "TAQUARUCU",
            "ROSANA",
            "MAUA",
            "STA CLARA PR",
            "FUNDAO",
            "G.B. MUNHOZ",
            "BAIXO IGUACU",
            "BARRA GRANDE",
            "GARIBALDI",
            "CAMPOS NOVOS",
            "MACHADINHO",
            "ITA",
            "PASSO FUNDO",
            "MONJOLINHO",
            "QUEBRA QUEIX",
            "SAO JOSE",
            "PASSO S JOAO",
            "FOZ CHAPECO",
            "CASTRO ALVES",
            "MONTE CLARO",
            "14 DE JULHO",
            "ERNESTINA",
            "PASSO REAL",
            "JACUI",
            "ITAUBA",
            "D. FRANCISCA",
            "G.P. SOUZA",
            "SALTO PILAO",
            "JORDAO",
            "SEGREDO",
            "SLT.SANTIAGO",
            "SALTO OSORIO",
            "SALTO CAXIAS",
            "SOBRADINHO",
            "ITAPARICA",
            "COMP PAF-MOX",
            "XINGO",
            "FICT.QUEIMAD",
            "FICT.TRES MA",
            "FICT.RETIRO",
            "ITAPEBI",
            "FICT.IRAPE",
            "P. CAVALO",
            "B. ESPERANCA",
            "CACH.CALDEIR",
            "ESTREITO TOC",
            "TUCURUI",
            "BALBINA",
            "COARACY NUNE",
            "FERREIRA GOM",
            "STO ANT JARI",
            "FICT.SERRA M",
            "FICT.CANA BR",
            "FICT.LAJEADO",
            "FICT.PEIXE A",
            "FICT.SAO SAL",
            "PIMENTAL",
            "CURUA-UNA",
            "BELO MONTE",
        ],
    )

    assert f.mercado_liquido.shape == (8, 4)
    assert f.energia_armazenada_absoluta_inicial.shape == (24, 4)
    assert f.energia_natural_afluente.shape == (24, 4)
    assert f.geracao_hidraulica_controlavel.shape == (72, 5)
    assert f.energia_vertida.shape == (24, 4)
    assert f.energia_armazenada_absoluta_final.shape == (24, 4)
    assert f.energia_natural_afluente_fio_bruta.shape == (24, 4)
    assert f.energia_evaporada.shape == (24, 4)
    assert f.energia_enchimento_volume_morto.shape == (24, 4)
    assert f.geracao_termica.shape == (720, 5)
    assert f.deficit.shape == (24, 6)
    assert f.valor_agua.shape == (24, 4)
    assert f.custo_marginal_operacao.shape == (24, 5)
    assert f.geracao_hidraulica_fio_liquida.shape == (24, 4)
    assert f.perdas_geracao_hidraulica_fio.shape == (24, 4)
    assert f.intercambio.shape == (120, 6)
    assert f.excesso.shape == (24, 5)
    assert f.energia_natural_afluente_bruta.shape == (24, 4)
    assert f.energia_natural_afluente_controlavel_corrigida.shape == (24, 4)
    assert f.geracao_hidraulica_maxima.shape == (72, 5)
    assert f.energia_afluente_controlavel_desvio.shape == (24, 4)
    assert f.energia_afluente_fio_desvio.shape == (24, 4)
    assert f.beneficio_intercambio.shape == (120, 6)
    assert f.fator_correcao_energia_natural_afluente_controlavel.shape == (
        24,
        4,
    )
    assert f.violacao_curva_aversao.shape == (24, 4)
    assert f.acionamento_curva_aversao.shape == (24, 4)
    assert f.penalidade_curva_aversao.shape == (24, 4)
    assert f.custo_operacao.shape == (2, 3)
    assert f.custo_geracao_termica.shape == (8, 4)
    assert f.beneficio_agrupamento_intercambio.shape == (36, 5)
    assert f.energia_natural_afluente_fio.shape == (24, 4)
    assert f.beneficio_despacho_gnl.shape == (48, 6)
    assert f.violacao_geracao_hidraulica_minima.shape == (72, 5)
    assert f.violacao_energia_vazao_minima.shape == (24, 4)
    assert (
        f.geracao_hidraulica_maxima_considerando_restricoes_eletricas.shape
        == (72, 5)
    )
    assert f.volume_armazenado_absoluto_final.shape == (324, 4)
    assert f.geracao_hidraulica_usina.shape == (972, 5)
    assert f.volume_turbinado.shape == (972, 5)
    assert f.volume_vertido.shape == (972, 5)
    assert f.violacao_geracao_hidraulica_minima_usina.shape == (972, 5)
    assert f.enchimento_volume_morto_usina.shape == (324, 4)
    assert f.violacao_defluencia_minima.shape == (972, 5)
    assert f.volume_desvio_usina.shape == (324, 4)
    assert f.volume_desvio_positivo_usina.shape == (324, 4)
    assert f.volume_desvio_negativo_usina.shape == (324, 4)
    assert f.violacao_fpha.shape == (972, 5)
    assert f.vazao_afluente.shape == (324, 4)
    assert f.vazao_incremental.shape == (324, 4)
    assert f.volume_armazenado_percentual_final.shape == (324, 4)
    assert f.custo_violacao_energia_vazao_minima.shape == (24, 4)
    assert f.custo_energia_afluente_controlavel_desvio.shape == (24, 4)
    assert f.custo_energia_afluente_fio_desvio.shape == (24, 4)
    assert f.custo_violacao_geracao_hidraulica_minima.shape == (72, 5)
    assert f.geracao_eolica.shape == (0, 5)
    assert f.velocidade_vento.shape == (0, 4)
    assert f.violacao_funcao_producao_eolica.shape == (0, 5)
    assert f.violacao_defluencia_maxima.shape == (972, 5)
    assert f.violacao_turbinamento_maximo.shape == (972, 5)
    assert f.violacao_turbinamento_minimo.shape == (972, 5)
    assert f.violacao_lpp_turbinamento_maximo.shape == (72, 5)
    assert f.violacao_lpp_defluencia_maxima.shape == (72, 5)
    assert f.violacao_lpp_turbinamento_maximo_usina.shape == (972, 5)
    assert f.violacao_lpp_defluencia_maxima_usina.shape == (972, 5)
    assert f.rhs_lpp_turbinamento_maximo.shape == (72, 5)
    assert f.rhs_lpp_defluencia_maxima.shape == (72, 5)
    assert f.rhs_lpp_turbinamento_maximo_usina.shape == (972, 5)
    assert f.rhs_lpp_defluencia_maxima_usina.shape == (972, 5)
    assert f.violacao_restricoes_eletricas_especiais.shape == (0, 5)
    assert f.custo_restricoes_eletricas_especiais.shape == (0, 5)
    assert f.volume_armazenado_absoluto_inicial.shape == (324, 4)
    assert f.valor_agua_usina.shape == (324, 4)
    assert f.volume_evaporado.shape == (324, 4)
    assert f.volume_bombeado.shape == (0, 5)
    assert f.consumo_energia_estacao_bombeamento.shape == (0, 5)
    assert f.volume_canal_desvio_usina.shape == (972, 5)



from inewave.nwlistop.modelos.blocos.submercado import Submercado
from inewave.nwlistop.modelos.merclsin import MerclAnos

from inewave.nwlistop.modelos.arquivos.arquivosubmercado import (
    ArquivoSubmercado,
)


class Mercl(ArquivoSubmercado):
    """
    Armazena os dados das saídas referentes ao mercado líquido
    de cada estágio em cada série por submercado.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `mercl001.out`.
    """

    BLOCKS = [
        Submercado,
        MerclAnos,
    ]



from inewave.newave.evap_avl_desv import EvapAvlDesv

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.nwv_avl_evap import MockNwvAvlEvap

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_atributos_encontrados_evap_avl_desv():
    m: MagicMock = mock_open(read_data="".join(MockNwvAvlEvap))
    with patch("builtins.open", m):
        rel = EvapAvlDesv.read(ARQ_TESTE)
        assert rel.versao == "FPHA_NEWAVE"
        assert rel.tabela.at[0, "periodo"] == 1
        assert rel.tabela.at[0, "codigo_usina"] == 4
        assert rel.tabela.at[0, "nome_usina"] == "FUNIL-GRANDE"
        assert rel.tabela.at[0, "volume_armazenado_hm3"] == 265.86
        assert rel.tabela.at[0, "evaporacao_calculada_hm3"] == 0.23
        assert rel.tabela.at[0, "evaporacao_modelo_hm3"] == 0.23
        assert rel.tabela.at[0, "desvio_absoluto_hm3"] == 0.0
        assert rel.tabela.at[0, "desvio_percentual"] == 0.0


def test_eq_evap_avl_desv():
    m: MagicMock = mock_open(read_data="".join(MockNwvAvlEvap))
    with patch("builtins.open", m):
        rel1 = EvapAvlDesv.read(ARQ_TESTE)
        rel2 = EvapAvlDesv.read(ARQ_TESTE)
        assert rel1 == rel2


def test_neq_evap_avl_desv():
    m: MagicMock = mock_open(read_data="".join(MockNwvAvlEvap))
    with patch("builtins.open", m):
        rel1 = EvapAvlDesv.read(ARQ_TESTE)
        rel2 = EvapAvlDesv.read(ARQ_TESTE)
        rel1.tabela.iloc[0, 0] = -1
        assert rel1 != rel2



from inewave.newave.modelos.cadic import BlocoCargasAdicionais

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Cadic(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às cargas
    adicionais.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoCargasAdicionais]

    @property
    def cargas(self) -> Optional[pd.DataFrame]:
        """
        Tabela com as cargas adicionais por mês/ano e por subsistema
        para cada razão de carga adicional. As colunas são:

        - codigo_submercado (`int`)
        - nome_submercado (`str`)
        - razao (`str`)
        - data (`datetime`)
        - valor (`float`)

        **OBS:** Na existência de períodos PRE ou POS, são adotados os anos
        padrão "0001" para PRE e "9999" para POS.

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoCargasAdicionais)
        if isinstance(b, BlocoCargasAdicionais):
            return b.data
        return None

    @cargas.setter
    def cargas(self, valor: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoCargasAdicionais)
        if isinstance(b, BlocoCargasAdicionais):
            b.data = valor
        else:
            raise ValueError("Campo não lido")



"""
========================================
Alteração no dger.dat
========================================
"""


# %%
# O dger.dat é um dos principais arquivos de entrada do modelo NEWAVE. É
# comum desejar realizar alterações automáticas neste arquivo, para fins de
# estudos periódicos. Isto pode ser feito utilizando a classe `Dger`:
from inewave.newave import Dger

arq_dger = Dger.read("./newave/dger.dat")

# %%
# Uma alteração muito comum é o nome do caso:
arq_dger.nome_caso

# %%
# Para alterá-lo, basta fazer:
arq_dger.nome_caso = "Caso de teste - inewave"
arq_dger.nome_caso

# %%
# Todos os campos do arquivo são editáveis seguindo esta mesma lógica. Os nomes das propriedades
# que foram escolhidas para implementação buscam refletir o sentido da configuração. Por exemplo,
# para as opções do gerenciamento externo de PLs, existem as propriedades separadas:

print(
    arq_dger.utiliza_gerenciamento_pls,
    arq_dger.comunicacao_dois_niveis,
    arq_dger.armazenamento_local_arquivos_temporarios,
)


# %%
# Para exportar o arquivo modificado, basta utilizar o método `write` a partir do objeto que foi
# alterado. Todavia, não é obrigatório fornecer um caminho para um arquivo no disco. A exportação
# também pode ser feita para um buffer em memória, se o objetivo
# for enviar o conteúdo do arquivo através de uma requisição HTTP, por exemplo, ou armazenar em um
# banco de dados para documentos:
from io import StringIO

conteudo_dger = StringIO()
arq_dger.write(conteudo_dger)
print(conteudo_dger.getvalue())



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousinapatamar import (
    ArquivoUsinaPatamar,
)
from inewave.nwlistop.modelos.viol_ghminuh import ViolGhminuhAnos


class ViolGhminuh(ArquivoUsinaPatamar):
    """
    Armazena os dados das saídas referentes à violação da meta de
    geração hidráulica mínima por usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `viol_ghminuh00x.out`, onde x varia conforme a
    usina em questão.

    """

    BLOCKS = [
        Usina,
        ViolGhminuhAnos,
    ]



from inewave.newave.modelos.cvar import (
    BlocoValoresConstantesCVAR,
    BlocoAlfaVariavelNoTempo,
    BlocoLambdaVariavelNoTempo,
)

from cfinterface.files.blockfile import BlockFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Cvar(BlockFile):
    """
    Armazena os dados de entrada do NEWAVE referentes à curva para
    penalização por volume mínimo dos reservatórios.
    """

    T = TypeVar("T")

    BLOCKS = [
        BlocoValoresConstantesCVAR,
        BlocoAlfaVariavelNoTempo,
        BlocoLambdaVariavelNoTempo,
    ]

    @property
    def valores_constantes(self) -> Optional[list]:
        """
        Valores constantes dos parâmetros ALFA e LAMBDA do CVAR.

        :return: Os valores dos campos da linha como uma lista.
        :rtype: list | None
        """
        b = self.data.get_blocks_of_type(BlocoValoresConstantesCVAR)
        if isinstance(b, BlocoValoresConstantesCVAR):
            return b.data
        return None

    @valores_constantes.setter
    def valores_constantes(self, valores: list):
        """
        Valores constantes dos parâmetros ALFA e LAMBDA do CVAR.

        :return: Os valores dos campos da linha como uma lista.
        :rtype: list | None
        """
        b = self.data.get_blocks_of_type(BlocoValoresConstantesCVAR)
        if isinstance(b, BlocoValoresConstantesCVAR):
            b.data = valores
        else:
            raise ValueError("Bloco não lido")

    @property
    def alfa_variavel(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os valores variáveis do CVAR para o parâmetro ALFA.

        - data (`datetime`)
        - valor (`float`)

        **OBS:** Na existência de períodos PRE ou POS, são adotados os anos
        padrão "0001" para PRE e "9999" para POS.


        :return: O valor de ALFA por estágio em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoAlfaVariavelNoTempo)
        if isinstance(b, BlocoAlfaVariavelNoTempo):
            return b.data
        return None

    @property
    def lambda_variavel(self) -> Optional[pd.DataFrame]:
        """
        Tabela com os valores variáveis do CVAR para o parâmetro LAMBDA.

        - data (`datetime`)
        - valor (`float`)

        **OBS:** Na existência de períodos PRE ou POS, são adotados os anos
        padrão "0001" para PRE e "9999" para POS.


        :return: O valor de LAMBDA por estágio em um DataFrame.
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_blocks_of_type(BlocoLambdaVariavelNoTempo)
        if isinstance(b, BlocoLambdaVariavelNoTempo):
            return b.data
        return None



# Rotinas de testes associadas ao arquivo MEDIAS-UISH.CSV do NWLISTOP
from inewave.nwlistop.mediasusih import Mediasusih

ARQ_TESTE = "tests/_arquivos/MEDIAS-USIH.CSV"


def test_eq_mediasusih():
    leitor = Mediasusih.read(ARQ_TESTE)
    leitor2 = Mediasusih.read(ARQ_TESTE)
    assert leitor == leitor2
    assert leitor.valores.shape == (9840, 30)



from datetime import datetime
from inewave.libs.modelos.restricoes import (
    RegistroRE,
    RegistroREHorizPer,
    RegistroRELimFormPer,
    RegistroRHE,
    RegistroRHEHorizPer,
    RegistroRHELsLPPEarmi,
    RegistroRHQ,
    RegistroRHQHorizPer,
    RegistroRHQLsLPPVoli,
    RegistroRHQLimFormPerPat,
    RegistroRHV,
    RegistroRHVHorizPer,
    RegistroRHVLimFormPer,
)

from inewave.libs.restricoes import Restricoes

from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.restricoes import (
    MockRE,
    MockREHorizPer,
    MockRELimFormPer,
    MockRHE,
    MockRHEHorizPer,
    MockRHELsLPPEarmi,
    MockRHQ,
    MockRHQHorizPer,
    MockRHQLsLPPVoli,
    MockRHQLimFormPerPat,
    MockRHV,
    MockRHVHorizPer,
    MockRHVLimFormPer,
    MockRestricoes,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_registro_re():
    m: MagicMock = mock_open(read_data="".join(MockRE))
    r = RegistroRE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, "1.0ger_usit(13) + ger_usih(66)"]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 5
    assert r.formula == "1.0ger_usit(13) + ger_usih(66)"
    r.formula = "Teste"


def test_registro_re_horiz_per():
    m: MagicMock = mock_open(read_data="".join(MockREHorizPer))
    r = RegistroREHorizPer()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, datetime(2021, 1, 1), datetime(2021, 1, 1)]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 2
    assert r.data_inicio == datetime(2021, 1, 1)
    r.data_inicio = datetime(2021, 2, 1)
    assert r.data_fim == datetime(2021, 1, 1)
    r.data_fim = datetime(2022, 11, 1)


def test_registro_rhq_lim_form_per():
    m: MagicMock = mock_open(read_data="".join(MockRELimFormPer))
    r = RegistroRELimFormPer()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [
        1,
        datetime(2021, 1, 1),
        datetime(2021, 3, 1),
        1,
        "-1.1e30",
        "5000",
    ]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 2
    assert r.data_inicio == datetime(2021, 1, 1)
    r.data_inicio = datetime(2021, 2, 1)
    assert r.data_fim == datetime(2021, 3, 1)
    r.data_fim = datetime(2021, 4, 1)
    assert r.limite_inferior == "-1.1e30"
    r.limite_inferior = "0"
    assert r.limite_inferior == "0"
    assert r.limite_superior == "5000"
    r.limite_superior = "0"
    assert r.limite_superior == "0"


def test_registro_rhe():
    m: MagicMock = mock_open(read_data="".join(MockRHE))
    r = RegistroRHE()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, "ger_ree(1) + ener_ver_ree(1)"]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 5
    assert r.formula == "ger_ree(1) + ener_ver_ree(1)"
    r.formula = "Teste"


def test_registro_rhe_horiz_per():
    m: MagicMock = mock_open(read_data="".join(MockRHEHorizPer))
    r = RegistroRHEHorizPer()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [2, datetime(2021, 1, 1), datetime(2025, 12, 1)]
    assert r.codigo_restricao == 2
    r.codigo_restricao = 1
    assert r.data_inicio == datetime(2021, 1, 1)
    r.data_inicio = datetime(2021, 2, 1)
    assert r.data_fim == datetime(2025, 12, 1)
    r.data_fim = datetime(2025, 11, 1)


def test_registro_rhe_ls_lpp_earmi():
    m: MagicMock = mock_open(read_data="".join(MockRHELsLPPEarmi))
    r = RegistroRHELsLPPEarmi()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [2, 2, -0.3, 4000.0]
    assert r.codigo_restricao == 2
    r.codigo_restricao = 1
    assert r.indice_reta == 2
    r.indice_reta = 1
    assert r.coeficiente_angular == -0.3
    r.coeficiente_angular = -0.5
    assert r.coeficiente_linear == 4000.0
    r.coeficiente_linear = 3500.0


def test_registro_rhq():
    m: MagicMock = mock_open(read_data="".join(MockRHQ))
    r = RegistroRHQ()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, "qtur(66)"]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 5
    assert r.formula == "qtur(66)"
    r.formula = "Teste"


def test_registro_rhq_horiz_per():
    m: MagicMock = mock_open(read_data="".join(MockRHQHorizPer))
    r = RegistroRHQHorizPer()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, datetime(2021, 1, 1), datetime(2022, 12, 1)]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 2
    assert r.data_inicio == datetime(2021, 1, 1)
    r.data_inicio = datetime(2021, 2, 1)
    assert r.data_fim == datetime(2022, 12, 1)
    r.data_fim = datetime(2022, 11, 1)


def test_registro_rhq_ls_lpp_voli():
    m: MagicMock = mock_open(read_data="".join(MockRHQLsLPPVoli))
    r = RegistroRHQLsLPPVoli()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, 1, 0.3, 3000.0]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 2
    assert r.indice_reta == 1
    r.indice_reta = 2
    assert r.coeficiente_angular == 0.3
    r.coeficiente_angular = 0.5
    assert r.coeficiente_linear == 3000.0
    r.coeficiente_linear = 3500.0


def test_registro_rhq_lim_form_per_pat():
    m: MagicMock = mock_open(read_data="".join(MockRHQLimFormPerPat))
    r = RegistroRHQLimFormPerPat()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [
        7,
        datetime(2023, 3, 1),
        datetime(2023, 6, 1),
        1,
        1000.0,
        1500.0,
    ]
    assert r.codigo_restricao == 7
    r.codigo_restricao = 2
    assert r.data_inicio == datetime(2023, 3, 1)
    r.data_inicio = datetime(2023, 4, 1)
    assert r.data_fim == datetime(2023, 6, 1)
    r.data_fim = datetime(2023, 5, 1)
    assert r.patamar == 1
    r.patamar = 2
    assert r.limite_inferior == 1000.0
    r.limite_inferior = 3500.0
    assert r.limite_superior == 1500.0
    r.limite_superior = 3000.0


def test_registro_rhv():
    m: MagicMock = mock_open(read_data="".join(MockRHV))
    r = RegistroRHV()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)
    assert r.data == [1, "1*vtur(18) + 2*vver(17) + 2*varm(18)"]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 5
    assert r.formula == "1*vtur(18) + 2*vver(17) + 2*varm(18)"
    r.formula = "Teste"


def test_registro_rhv_horiz_per():
    m: MagicMock = mock_open(read_data="".join(MockRHVHorizPer))
    r = RegistroRHVHorizPer()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [1, datetime(2023, 3, 1), datetime(2023, 6, 1)]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 2
    assert r.data_inicio == datetime(2023, 3, 1)
    r.data_inicio = datetime(2021, 2, 1)
    assert r.data_fim == datetime(2023, 6, 1)
    r.data_fim = datetime(2022, 11, 1)


def test_registro_rhv_lim_form_per():
    m: MagicMock = mock_open(read_data="".join(MockRHVLimFormPer))
    r = RegistroRHVLimFormPer()
    with patch("builtins.open", m):
        with open("", "") as fp:
            r.read(fp)

    assert r.data == [
        1,
        datetime(2023, 3, 1),
        datetime(2023, 6, 1),
        100.0,
        100.0,
    ]
    assert r.codigo_restricao == 1
    r.codigo_restricao = 2
    assert r.data_inicio == datetime(2023, 3, 1)
    r.data_inicio = datetime(2023, 4, 1)
    assert r.data_fim == datetime(2023, 6, 1)
    r.data_fim = datetime(2023, 5, 1)
    assert r.limite_inferior == 100.0
    r.limite_inferior = 150.0
    assert r.limite_superior == 100.0
    r.limite_superior = 300.0


def test_atributos_encontrados_restricoes():
    m: MagicMock = mock_open(read_data="".join(MockRestricoes))
    with patch("builtins.open", m):
        e = Restricoes.read(ARQ_TESTE)
        assert len(e.re()) > 0
        assert len(e.re_horiz_per()) > 0
        assert len(e.re_lim_form_per()) > 0
        assert len(e.rhe()) > 0
        assert len(e.rhe_horiz_per()) > 0
        assert len(e.rhe_ls_lpp_earmi()) > 0
        assert len(e.rhq()) > 0
        assert len(e.rhq_horiz_per()) > 0
        assert len(e.rhq_ls_lpp_voli()) > 0
        assert len(e.rhq_lim_form_per_pat()) > 0
        assert len(e.rhv()) > 0
        assert len(e.rhv_horiz_per()) > 0
        assert len(e.rhv_lim_form_per()) > 0


def test_eq_restricoes():
    m: MagicMock = mock_open(read_data="".join(MockRestricoes))
    with patch("builtins.open", m):
        cf1 = Restricoes.read(ARQ_TESTE)
        cf2 = Restricoes.read(ARQ_TESTE)
        assert cf1 == cf2


def test_neq_restricoes():
    m: MagicMock = mock_open(read_data="".join(MockRestricoes))
    with patch("builtins.open", m):
        cf1 = Restricoes.read(ARQ_TESTE)
        cf2 = Restricoes.read(ARQ_TESTE)
        cf2.data.remove(cf1.re()[0])
        assert cf1 != cf2


def test_leitura_escrita_restricoes():
    m_leitura: MagicMock = mock_open(read_data="".join(MockRestricoes))
    with patch("builtins.open", m_leitura):
        cf1 = Restricoes.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = Restricoes.read(ARQ_TESTE)
        assert cf1 == cf2



from inewave.newave.modelos.adterm import BlocoUTEAdTerm

from cfinterface.files.sectionfile import SectionFile
from typing import TypeVar, Optional
import pandas as pd  # type: ignore


class Adterm(SectionFile):
    """
    Armazena os dados de entrada do NEWAVE referentes às térmicas de
    despacho antecipado disponíveis.

    """

    T = TypeVar("T")

    SECTIONS = [BlocoUTEAdTerm]

    @property
    def despachos(self) -> Optional[pd.DataFrame]:
        """
        A tabela de espachos antecipados das térmicas GNL.

        - codigo_usina (`int`)
        - nome_usina (`str`)
        - lag (`int`)
        - patamar (`int`)
        - valor (`float`)

        :return: A tabela como um DataFrame
        :rtype: pd.DataFrame | None
        """
        b = self.data.get_sections_of_type(BlocoUTEAdTerm)
        if isinstance(b, BlocoUTEAdTerm):
            return b.data
        return None

    @despachos.setter
    def despachos(self, d: pd.DataFrame):
        b = self.data.get_sections_of_type(BlocoUTEAdTerm)
        if isinstance(b, BlocoUTEAdTerm):
            b.data = d



from inewave.nwlistop.modelos.arquivos.arquivosinpatamar import (
    ArquivoSINPatamar,
)
from inewave.nwlistop.modelos.dlppdfmaxs import DLPPdfmaxAnos


class Dlppdfmaxs(ArquivoSINPatamar):
    """
    Armazena os dados das saídas referentes à violação das restrições
    LPP de defluência máxima por patamar para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `dlppdfmaxs.out`.
    """

    BLOCKS = [
        DLPPdfmaxAnos,
    ]



# Rotinas de testes associadas ao arquivo shist.dat do NEWAVE
from inewave.newave.modelos.shist import (
    BlocoVarreduraShist,
    BlocoSeriesSimulacaoShist,
)

from inewave.newave import Shist


from tests.mocks.mock_open import mock_open
from unittest.mock import MagicMock, patch

from tests.mocks.arquivos.shist import (
    MockBlocoVarreduraShist,
    MockBlocoSeriesSimulacaoShist,
    MockShist,
)

ARQ_TESTE = "./tests/mocks/arquivos/__init__.py"


def test_bloco_varredura_shist():
    m: MagicMock = mock_open(read_data="".join(MockBlocoVarreduraShist))
    b = BlocoVarreduraShist()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == [1, 1932]


def test_bloco_series_shist():
    m: MagicMock = mock_open(read_data="".join(MockBlocoSeriesSimulacaoShist))
    b = BlocoSeriesSimulacaoShist()
    with patch("builtins.open", m):
        with open("", "") as fp:
            b.read(fp)

    assert b.data == []


def test_atributos_encontrados_shist():
    m: MagicMock = mock_open(read_data="".join(MockShist))
    with patch("builtins.open", m):
        ad = Shist.read(ARQ_TESTE)
        assert ad.varredura is not None
        assert ad.ano_inicio_varredura is not None
        assert ad.anos_inicio_simulacoes is not None


def test_atributos_nao_encontrados_shist():
    m: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m):
        ad = Shist.read(ARQ_TESTE)
        assert ad.varredura is None
        assert ad.ano_inicio_varredura is None
        assert ad.anos_inicio_simulacoes == []


def test_eq_shist():
    m: MagicMock = mock_open(read_data="".join(MockShist))
    with patch("builtins.open", m):
        cf1 = Shist.read(ARQ_TESTE)
        cf2 = Shist.read(ARQ_TESTE)
        assert cf1 == cf2


def test_neq_shist():
    m: MagicMock = mock_open(read_data="".join(MockShist))
    with patch("builtins.open", m):
        cf1 = Shist.read(ARQ_TESTE)
        cf2 = Shist.read(ARQ_TESTE)
        cf2.varredura = 0
        assert cf1 != cf2


def test_leitura_escrita_shist():
    m_leitura: MagicMock = mock_open(read_data="".join(MockShist))
    with patch("builtins.open", m_leitura):
        cf1 = Shist.read(ARQ_TESTE)
    m_escrita: MagicMock = mock_open(read_data="")
    with patch("builtins.open", m_escrita):
        cf1.write(ARQ_TESTE)
        # Recupera o que foi escrito
        chamadas = m_escrita.mock_calls
        linhas_escritas = [
            chamadas[i].args[0] for i in range(1, len(chamadas) - 1)
        ]
    m_releitura: MagicMock = mock_open(read_data="".join(linhas_escritas))
    with patch("builtins.open", m_releitura):
        cf2 = Shist.read(ARQ_TESTE)
        assert cf1 == cf2



from inewave.nwlistop.modelos.arquivos.arquivosin import (
    ArquivoSIN,
)
from inewave.nwlistop.modelos.viol_evminsin import ViolEvminsinAnos


class ViolEvminsin(ArquivoSIN):
    """
    Armazena os dados das saídas referentes às violações da meta
    de energia da vazão mínima para o SIN.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `viol_evminsin.out`

    """

    BLOCKS = [
        ViolEvminsinAnos,
    ]



from cfinterface.components.register import Register
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from cfinterface.adapters.components.repository import factory
from copy import deepcopy
from datetime import datetime
from typing import Optional, IO, List


class ModifRegister(Register):
    """
    Registro base para o arquivo modif.dat, que contém
    características específicas deste arquivo:

    - Possibilidade de ler uma linha assumindo que o conteúdo
        de cada campo é separado por um número variável
        de espaços.
    - Capacidade de escrever uma linha assumindo uma formatação
        constante para melhor visualização.

    OBS: Atualmente só utilizados para registros VOLMIN e VOLMAX.
    """

    __slots__ = ["__identifier_field", "__previous", "__next", "__data"]

    def __init__(
        self,
        previous=None,
        next=None,
        data=None,
    ) -> None:
        self.__identifier_field: Field = LiteralField(
            self.__class__.IDENTIFIER_DIGITS, 0
        )
        self.__previous = previous
        self.__next = next
        if data is None:
            self.__data = [None] * len(self.__class__.LINE.fields)
        else:
            self.__data = data
        super().__init__(previous, next, data)

    def read(self, file: IO, storage: str = "", *args, **kwargs) -> bool:
        delimited_fields = [deepcopy(f) for f in self.__class__.LINE.fields]
        line = Line(
            [self.__identifier_field] + delimited_fields,
            delimiter=" ",
            storage=storage,
        )
        line_full_spaces = file.readline().strip()
        line_parts = [p for p in line_full_spaces.split(" ") if len(p) > 0]
        line_simple_spaces = " ".join(line_parts)
        self.data = line.read(line_simple_spaces)[1:]
        return True

    def write(self, file: IO, storage: str = "", *args, **kwargs) -> bool:
        if not self.empty:
            line = Line(
                [self.__identifier_field] + self.__class__.LINE.fields,
                delimiter=self.__class__.LINE.delimiter,
                storage=storage,
            )
            linedata = line.write([self.__class__.IDENTIFIER] + self.data)
            factory(storage).write(file, linedata)
        return True


class USINA(Register):
    """
    Registro que contém a usina modificada.
    """

    __slots__: List[str] = []

    IDENTIFIER = " USINA"
    IDENTIFIER_DIGITS = 8
    LINE = Line([IntegerField(4, 9), LiteralField(20, 44)])

    @property
    def codigo(self) -> Optional[int]:
        """
        O principal conteúdo do registro (código da usina).

        :return: O código da usina
        :rtype: Optional[int]
        """
        return self.data[0]

    @codigo.setter
    def codigo(self, t: int):
        self.data[0] = t

    @property
    def nome(self) -> Optional[str]:
        """
        O nome da usina (opcional).

        :return: O nome da usina
        :rtype: Optional[str]
        """
        return self.data[1]

    @nome.setter
    def nome(self, t: str):
        self.data[1] = t


class VOLMIN(ModifRegister):
    """
    Registro que contém uma modificação de volume mínimo
    operativo para uma usina.
    """

    __slots__ = []

    IDENTIFIER = " VOLMIN"
    IDENTIFIER_DIGITS = 8
    LINE = Line([FloatField(8, 10, 2), LiteralField(3, 19)])

    @property
    def volume(self) -> Optional[float]:
        """
        O novo valor de volume

        :return: O novo valor de volume
        :rtype: Optional[float]
        """
        return self.data[0]

    @volume.setter
    def volume(self, t: float):
        self.data[0] = t

    @property
    def unidade(self) -> Optional[str]:
        """
        A unidade do volume informado

        :return: A unidade do volume
        :rtype: Optional[str]
        """
        return self.data[1]

    @unidade.setter
    def unidade(self, t: str):
        self.data[1] = t


class VOLMAX(ModifRegister):
    """
    Registro que contém uma modificação de volume máximo
    operativo para uma usina.
    """

    __slots__ = []

    IDENTIFIER = " VOLMAX"
    IDENTIFIER_DIGITS = 8
    LINE = Line([FloatField(6, 10, 3), LiteralField(3, 17)])

    @property
    def volume(self) -> Optional[float]:
        """
        O novo valor de volume

        :return: O novo valor de volume
        :rtype: Optional[float]
        """
        return self.data[0]

    @volume.setter
    def volume(self, t: float):
        self.data[0] = t

    @property
    def unidade(self) -> Optional[str]:
        """
        A unidade do volume informado

        :return: A unidade do volume
        :rtype: Optional[str]
        """
        return self.data[1]

    @unidade.setter
    def unidade(self, t: str):
        self.data[1] = t


class NUMCNJ(ModifRegister):
    """
    Registro que contém uma modificação de número de conjunto
    de máquinas.
    """

    __slots__ = []

    IDENTIFIER = " NUMCNJ"
    IDENTIFIER_DIGITS = 8
    LINE = Line([IntegerField(2, 11)])

    @property
    def numero(self) -> int:
        """
        O novo valor do número de conjuntos

        :return: O novo número de conjuntos
        :rtype: Optional[int]
        """
        return self.data[0]

    @numero.setter
    def numero(self, t: int):
        self.data[0] = t


class NUMMAQ(ModifRegister):
    """
    Registro que contém uma modificação do número de máquinas em um
    conjunto de máquinas.
    """

    __slots__ = []

    IDENTIFIER = " NUMMAQ"
    IDENTIFIER_DIGITS = 8
    LINE = Line([IntegerField(3, 11), IntegerField(3, 14)])

    @property
    def conjunto(self) -> Optional[int]:
        """
        O conjunto de máquinas que terá o número alterado

        :return: O índice do conjunto de máquinas
        :rtype: Optional[int]
        """
        return self.data[0]

    @conjunto.setter
    def conjunto(self, t: int):
        self.data[0] = t

    @property
    def numero_maquinas(self) -> Optional[int]:
        """
        O novo número de máquinas do conjunto

        :return: O número de máquinas do conjunto
        :rtype: Optional[int]
        """
        return self.data[1]

    @numero_maquinas.setter
    def numero_maquinas(self, t: int):
        self.data[1] = t


class VAZMIN(ModifRegister):
    """
    Registro que contém uma modificação de vazão mínima (m3/s).
    """

    __slots__ = []

    IDENTIFIER = " VAZMIN "
    IDENTIFIER_DIGITS = 8
    LINE = Line([IntegerField(8, 10)])

    @property
    def vazao(self) -> Optional[int]:
        """
        O valor de vazão mínima

        :return: A nova vazão
        :rtype: Optional[int]
        """
        return self.data[0]

    @vazao.setter
    def vazao(self, t: int):
        self.data[0] = t


class CFUGA(ModifRegister):
    """
    Registro que contém uma modificação do nível do canal de fuga.
    """

    __slots__ = []

    IDENTIFIER = " CFUGA"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [IntegerField(2, 10), IntegerField(4, 13), FloatField(7, 18, 3)]
    )

    @property
    def data_inicio(self) -> datetime:
        """
        A data de início da modificação

        :return: A data de início da modificação
        :rtype: Optional[datetime]
        """
        return datetime(self.data[1], self.data[0], 1)

    @data_inicio.setter
    def data_inicio(self, t: datetime):
        self.data[0] = t.month
        self.data[1] = t.year

    @property
    def nivel(self) -> float:
        """
        O novo nivel do canal de fuga

        :return: O novo nível
        :rtype: Optional[int]
        """
        return self.data[2]

    @nivel.setter
    def nivel(self, t: float):
        self.data[2] = t


class CMONT(ModifRegister):
    """
    Registro que contém uma modificação do nível do canal de montante.
    """

    __slots__ = []

    IDENTIFIER = " CMONT"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [IntegerField(2, 10), IntegerField(4, 13), FloatField(7, 18, 3)]
    )

    @property
    def data_inicio(self) -> datetime:
        """
        A data de início da modificação

        :return: A data de início da modificação
        :rtype: Optional[datetime]
        """
        return datetime(self.data[1], self.data[0], 1)

    @data_inicio.setter
    def data_inicio(self, t: datetime):
        self.data[0] = t.month
        self.data[1] = t.year

    @property
    def nivel(self) -> Optional[float]:
        """
        O novo nivel do canal de montante

        :return: O novo nível
        :rtype: Optional[float]
        """
        return self.data[2]

    @nivel.setter
    def nivel(self, t: float):
        self.data[2] = t


class VMAXT(Register):
    """
    Registro que contém uma modificação do volume máximo
    com data.
    """

    __slots__: List[str] = []

    IDENTIFIER = " VMAXT"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 3),
            LiteralField(3, 26),
        ]
    )

    @property
    def data_inicio(self) -> datetime:
        """
        A data de início da modificação

        :return: A data de início da modificação
        :rtype: Optional[datetime]
        """
        return datetime(self.data[1], self.data[0], 1)

    @data_inicio.setter
    def data_inicio(self, t: datetime):
        self.data[0] = t.month
        self.data[1] = t.year

    @property
    def volume(self) -> Optional[float]:
        """
        O novo volume máximo

        :return: O novo volume
        :rtype: Optional[float]
        """
        return self.data[2]

    @volume.setter
    def volume(self, t: float):
        self.data[2] = t

    @property
    def unidade(self) -> Optional[str]:
        """
        A unidade de fornecimento do volume

        :return: A unidade
        :rtype: Optional[str]
        """
        return self.data[3]

    @unidade.setter
    def unidade(self, t: str):
        self.data[3] = t


class VMINT(Register):
    """
    Registro que contém uma modificação do volume mínimo
    com data.
    """

    __slots__: List[str] = []

    IDENTIFIER = " VMINT"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 3),
            LiteralField(3, 26),
        ]
    )

    @property
    def data_inicio(self) -> datetime:
        """
        A data de início da modificação

        :return: A data de início da modificação
        :rtype: Optional[datetime]
        """
        return datetime(self.data[1], self.data[0], 1)

    @data_inicio.setter
    def data_inicio(self, t: datetime):
        self.data[0] = t.month
        self.data[1] = t.year

    @property
    def volume(self) -> Optional[float]:
        """
        O novo volume mínimo

        :return: O novo volume
        :rtype: Optional[float]
        """
        return self.data[2]

    @volume.setter
    def volume(self, t: float):
        self.data[2] = t

    @property
    def unidade(self) -> Optional[str]:
        """
        A unidade de fornecimento do volume

        :return: A unidade
        :rtype: Optional[str]
        """
        return self.data[3]

    @unidade.setter
    def unidade(self, t: str):
        self.data[3] = t


class VMINP(Register):
    """
    Registro que contém uma modificação do volume mínimo
    com data para adoção de penalidade.
    """

    __slots__: List[str] = []

    IDENTIFIER = " VMINP"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 3),
            LiteralField(3, 26),
        ]
    )

    @property
    def data_inicio(self) -> datetime:
        """
        A data de início da modificação

        :return: A data de início da modificação
        :rtype: Optional[datetime]
        """
        return datetime(self.data[1], self.data[0], 1)

    @data_inicio.setter
    def data_inicio(self, t: datetime):
        self.data[0] = t.month
        self.data[1] = t.year

    @property
    def volume(self) -> Optional[float]:
        """
        O novo volume mínimo a partir da data

        :return: O novo volume
        :rtype: Optional[float]
        """
        return self.data[2]

    @volume.setter
    def volume(self, t: float):
        self.data[2] = t

    @property
    def unidade(self) -> Optional[str]:
        """
        A unidade do volume fornecido

        :return: A unidade
        :rtype: Optional[str]
        """
        return self.data[3]

    @unidade.setter
    def unidade(self, t: str):
        self.data[3] = t


class VAZMINT(Register):
    """
    Registro que contém uma modificação da vazão mínima
    com data.
    """

    __slots__: List[str] = []

    IDENTIFIER = " VAZMINT"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 2),
        ]
    )

    @property
    def data_inicio(self) -> datetime:
        """
        A data de início da modificação

        :return: A data de início da modificação
        :rtype: Optional[datetime]
        """
        return datetime(self.data[1], self.data[0], 1)

    @data_inicio.setter
    def data_inicio(self, t: datetime):
        self.data[0] = t.month
        self.data[1] = t.year

    @property
    def vazao(self) -> Optional[float]:
        """
        A nova vazão mínima a partir da data

        :return: A nova vazão
        :rtype: Optional[float]
        """
        return self.data[2]

    @vazao.setter
    def vazao(self, t: float):
        self.data[2] = t


class VAZMAXT(Register):
    """
    Registro que contém uma modificação da vazão máxima
    com data.
    """

    __slots__: List[str] = []

    IDENTIFIER = " VAZMAXT"
    IDENTIFIER_DIGITS = 8
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 2),
        ]
    )

    @property
    def data_inicio(self) -> datetime:
        """
        A data de início da modificação

        :return: A data de início da modificação
        :rtype: Optional[datetime]
        """
        return datetime(self.data[1], self.data[0], 1)

    @data_inicio.setter
    def data_inicio(self, t: datetime):
        self.data[0] = t.month
        self.data[1] = t.year

    @property
    def vazao(self) -> Optional[float]:
        """
        A nova vazão máxima a partir da data

        :return: A nova vazão
        :rtype: Optional[float]
        """
        return self.data[2]

    @vazao.setter
    def vazao(self, t: float):
        self.data[2] = t


class TURBMAXT(Register):
    """
    Registro que contém uma modificação da turbinamento máximo
    com data.
    """

    __slots__: List[str] = []

    IDENTIFIER = " TURBMAXT"
    IDENTIFIER_DIGITS = 9
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 2),
        ]
    )

    @property
    def data_inicio(self) -> datetime:
        """
        A data de início da modificação

        :return: A data de início da modificação
        :rtype: Optional[datetime]
        """
        return datetime(self.data[1], self.data[0], 1)

    @data_inicio.setter
    def data_inicio(self, t: datetime):
        self.data[0] = t.month
        self.data[1] = t.year

    @property
    def turbinamento(self) -> Optional[float]:
        """
        O novo turbinamento máximo a partir da data

        :return: O novo turbinamento
        :rtype: Optional[float]
        """
        return self.data[2]

    @turbinamento.setter
    def turbinamento(self, t: float):
        self.data[2] = t


class TURBMINT(Register):
    """
    Registro que contém uma modificação da turbinamento mínimo
    com data.
    """

    __slots__: List[str] = []

    IDENTIFIER = " TURBMINT"
    IDENTIFIER_DIGITS = 9
    LINE = Line(
        [
            IntegerField(2, 10),
            IntegerField(4, 13),
            FloatField(7, 18, 2),
        ]
    )

    @property
    def data_inicio(self) -> datetime:
        """
        A data de início da modificação

        :return: A data de início da modificação
        :rtype: Optional[datetime]
        """
        return datetime(self.data[1], self.data[0], 1)

    @data_inicio.setter
    def data_inicio(self, t: datetime):
        self.data[0] = t.month
        self.data[1] = t.year

    @property
    def turbinamento(self) -> Optional[float]:
        """
        O novo turbinamento mínimo a partir da data

        :return: O novo turbinamento
        :rtype: Optional[float]
        """
        return self.data[2]

    @turbinamento.setter
    def turbinamento(self, t: float):
        self.data[2] = t



# Imports do próprio módulo

from inewave.config import (
    MAX_ANOS_ESTUDO,
    MAX_ANOS_HISTORICO,
    MAX_UHES,
    MESES_ABREV,
    MESES_DF,
)

# Imports de módulos externos
from cfinterface.components.block import Block
from cfinterface.components.line import Line
from cfinterface.components.field import Field
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.floatfield import FloatField
from datetime import date
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from typing import IO, List
from inewave._utils.formatacao import (
    prepara_vetor_anos_tabela,
    repete_vetor,
)


class BlocoSerieVazoesUHE(Block):
    """
    Bloco de informações do arquivo `parpvaz.dat`
    relativo às séries de vazões por usina.
    """

    __slots__ = [
        "__campo_uhe",
        "__campo_cfg",
        "__linha",
    ]

    BEGIN_PATTERN = "SERIE  DE VAZOES   DA USINA"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__campo_uhe = LiteralField(14, 57)
        self.__campo_cfg = IntegerField(5, 88)
        campo_ano: List[Field] = [IntegerField(4, 0)]
        campo_vazao: List[Field] = [
            FloatField(10, 4 + 11 * i, 2) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + campo_vazao)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoSerieVazoesUHE):
            return False
        bloco: BlocoSerieVazoesUHE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "data": prepara_vetor_anos_tabela(anos),
                    "valor": tabela.flatten(),
                }
            )
            df["uhe"] = uhe
            df["configuracao"] = cfg
            return df[["uhe", "configuracao", "data", "valor"]]

        # Identifica a usina e a configuração
        linha = file.readline()
        uhe = self.__campo_uhe.read(linha)
        cfg = self.__campo_cfg.read(linha)

        # Salta as linhas adicionais
        for _ in range(4):
            file.readline()

        # Variáveis auxiliares
        anos: List[int] = []
        tabela = np.zeros((MAX_ANOS_HISTORICO, len(MESES_DF)))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            anos.append(dados[0])
            tabela[i, :] = dados[1:]
            i += 1


class BlocoCorrelVazoesUHE(Block):
    """
    Bloco de informações do arquivo `parpvaz.dat`
    relativo às séries de autocorrelações
    das vazões por usina.
    """

    __slots__ = [
        "__linha",
    ]

    BEGIN_PATTERN = "CORRELOGRAMO DA SERIE DE VAZOES"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_mes: List[Field] = [LiteralField(3, 1)]
        campo_ano: List[Field] = [LiteralField(4, 5)]
        correl: List[Field] = [
            FloatField(10, 9 + 10 * i, 5) for i in range(11)
        ]
        self.__linha = Line(campo_mes + campo_ano + correl)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelVazoesUHE):
            return False
        bloco: BlocoCorrelVazoesUHE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_vetor_anos(anos: List[str]) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"]) // 12
            numero_anos_pos = len([p for p in anos if p == "POS"]) // 12
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # Descobre o primeiro ano de estudo
            # TODO - remover fallback de 1970 quando estiver o ano correto.
            primeiro_ano_estudo = (
                sorted(anos_estudo)[0] if len(anos_estudo) > 0 else 1970
            )
            # Descobre o último ano de estudo
            ultimo_ano_estudo = (
                sorted(anos_estudo)[-1] if len(anos_estudo) > 0 else 1970
            )
            indice_inicio_pos = anos.index("POS")
            # Substitui os anos pré e pós pelos valores específicos
            for a in range(numero_anos_pre):
                idx_i = 12 * a
                idx_f = idx_i + 12
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx_i:idx_f] = [str(ano)] * 12
            for a in range(numero_anos_pos):
                idx_i = indice_inicio_pos + 12 * a
                idx_f = idx_i + 12
                ano = ultimo_ano_estudo + a + 1
                anos[idx_i:idx_f] = [str(ano)] * 12
            return [int(a) for a in anos]

        def converte_vetor_meses(meses: List[str]) -> List[int]:
            return [MESES_ABREV.index(m) + 1 for m in meses]

        def converte_tabela_em_df():
            anos_conv = converte_vetor_anos(anos)
            meses_conv = converte_vetor_meses(meses)
            datas = [
                date(year=a, month=m, day=1)
                for a, m in zip(anos_conv, meses_conv)
            ]
            df = pd.DataFrame(
                data={
                    "data": repete_vetor(datas, len(MESES_DF) - 1),
                    "lag": np.tile(np.arange(1, len(MESES_DF)), len(datas)),
                    "valor": tabela.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(5):
            file.readline()

        # Variáveis auxiliares
        meses: List[str] = []
        anos: List[str] = []
        tabela = np.zeros((MAX_ANOS_HISTORICO * len(MESES_DF), 11))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            meses.append(dados[0])
            anos.append(dados[1])
            tabela[i, :] = dados[2:]
            i += 1


class BlocoCorrelParcialVazoesUHE(Block):
    """
    Bloco de informações do arquivo `parpvaz.dat`
    relativo às séries de autocorrelações parciais
    das vazões por UHE.
    """

    __slots__ = [
        "__linha",
    ]

    BEGIN_PATTERN = "CORRELOGRAMO PARCIAL DA SERIE DE VAZOES"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_mes: List[Field] = [LiteralField(3, 1)]
        campo_ano: List[Field] = [LiteralField(4, 5)]
        correl: List[Field] = [
            FloatField(10, 9 + 10 * i, 5) for i in range(11)
        ]
        self.__linha = Line(campo_mes + campo_ano + correl)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelParcialVazoesUHE):
            return False
        bloco: BlocoCorrelParcialVazoesUHE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_vetor_anos(anos: List[str]) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"]) // 12
            numero_anos_pos = len([p for p in anos if p == "POS"]) // 12
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # TODO - remover fallback de 1970 quando estiver o ano correto.
            # Descobre o primeiro ano de estudo
            primeiro_ano_estudo = (
                sorted(anos_estudo)[0] if len(anos_estudo) > 0 else 1970
            )
            # Descobre o último ano de estudo
            ultimo_ano_estudo = (
                sorted(anos_estudo)[-1] if len(anos_estudo) > 0 else 1970
            )
            indice_inicio_pos = anos.index("POS")
            # Substitui os anos pré e pós pelos valores específicos
            for a in range(numero_anos_pre):
                idx_i = 12 * a
                idx_f = idx_i + 12
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx_i:idx_f] = [str(ano)] * 12
            for a in range(numero_anos_pos):
                idx_i = indice_inicio_pos + 12 * a
                idx_f = idx_i + 12
                ano = ultimo_ano_estudo + a + 1
                anos[idx_i:idx_f] = [str(ano)] * 12
            return [int(a) for a in anos]

        def converte_vetor_meses(meses: List[str]) -> List[int]:
            return [MESES_ABREV.index(m) + 1 for m in meses]

        def converte_tabela_em_df():
            anos_conv = converte_vetor_anos(anos)
            meses_conv = converte_vetor_meses(meses)
            datas = [
                date(year=a, month=m, day=1)
                for a, m in zip(anos_conv, meses_conv)
            ]
            df = pd.DataFrame(
                data={
                    "data": repete_vetor(datas, len(MESES_DF) - 1),
                    "lag": np.tile(np.arange(1, len(MESES_DF)), len(datas)),
                    "valor": tabela.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(3):
            file.readline()

        # Variáveis auxiliares
        meses: List[str] = []
        anos: List[str] = []
        tabela = np.zeros((MAX_ANOS_HISTORICO * len(MESES_DF), 11))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            meses.append(dados[0])
            anos.append(dados[1])
            tabela[i, :] = dados[2:]
            i += 1


class BlocoOrdemModeloUHE(Block):
    """
    Bloco de informações do arquivo `parpvaz.dat`
    relativo às ordens finais ajustadas do modelo e aos
    seus coeficientes por UHE.
    """

    __slots__ = [
        "__linha",
        "__tipo",
    ]

    BEGIN_PATTERN = "DO MODELO AUTORREGRESSIVO PARA CADA PERIODO"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_ano: List[Field] = [LiteralField(4, 32)]
        orders: List[Field] = [
            IntegerField(5, 36 + 5 * i) for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campo_ano + orders)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoOrdemModeloUHE):
            return False
        bloco: BlocoOrdemModeloUHE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_vetor_anos(anos: List[str]) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"])
            numero_anos_pos = len([p for p in anos if p == "POS"])
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # TODO - remover fallback de 1970 quando estiver o ano correto.
            # Descobre o primeiro ano de estudo
            primeiro_ano_estudo = (
                sorted(anos_estudo)[0] if len(anos_estudo) > 0 else 1970
            )
            # Descobre o último ano de estudo
            ultimo_ano_estudo = (
                sorted(anos_estudo)[-1] if len(anos_estudo) > 0 else 1970
            )
            indice_inicio_pos = anos.index("POS")
            # Substitui os anos pré e pós pelos valores específicos
            for a in range(numero_anos_pre):
                idx = a
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx] = str(ano)
            for a in range(numero_anos_pos):
                idx = indice_inicio_pos + a
                ano = ultimo_ano_estudo + a + 1
                anos[idx] = str(ano)
            return [int(a) for a in anos]

        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "data": prepara_vetor_anos_tabela(
                        converte_vetor_anos(anos)
                    ),
                    "valor": tabela.flatten(),
                }
            )
            df["tipo"] = self.__tipo
            return df[["tipo", "data", "valor"]]

        # Salta as linhas adicionais
        linha = file.readline()
        self.__tipo = linha.split("ORDEM")[1].split("DO MODELO")[0].strip()
        for _ in range(3):
            file.readline()

        # Variáveis auxiliares
        anos: List[str] = []
        tabela = np.zeros((3 * MAX_ANOS_ESTUDO, len(MESES_DF)), dtype=np.int64)
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            anos.append(dados[0])
            tabela[i, :] = dados[1:]
            i += 1


class BlocoCoeficientesModeloUHE(Block):
    """
    Lista de coeficientes dos modelos PAR ou PAR-A por UHE.
    """

    __slots__ = [
        "__linha",
    ]

    BEGIN_PATTERN = " COEFICIENTES DA EQUACAO DE REGRESSAO DE UM PROCESSO"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        coefs: List[Field] = [
            FloatField(9, 11 * i, 3, format="E") for i in range(11)
        ]
        self.__linha = Line(coefs)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCoeficientesModeloUHE):
            return False
        bloco: BlocoCoeficientesModeloUHE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "tipo": repete_vetor(["psi", "psi_norm"], len(MESES_DF)),
                    "ordem": np.tile(list(range(1, len(MESES_DF))) + ["A"], 2),
                    "valor": tabela.flatten(),
                }
            )
            return df.dropna()

        linha = file.readline()

        tabela = np.zeros((1, 2 * len(MESES_DF)))
        # Processa os dados
        dados_psi = self.__linha.read(file.readline())
        dados_psi_norm = self.__linha.read(file.readline())
        linha = file.readline()
        if len(linha) < 3:
            dados_psi_a = [0.0, 0.0]
        else:
            dados_psi_a = [
                self.__linha.read(linha)[0],
                self.__linha.read(file.readline())[0],
            ]
        tabela[0, :] = (
            dados_psi + [dados_psi_a[0]] + dados_psi_norm + [dados_psi_a[1]]
        )
        self.data = converte_tabela_em_df()


class BlocoSerieRuidosUHE(Block):
    """
    Bloco de informações do arquivo `parpvaz.dat`
    relativo às séries de ruídos por usina.
    """

    __slots__ = [
        "__campo_ano",
        "__campo_cfg",
        "__linha",
    ]

    BEGIN_PATTERN = "SERIE DE RUIDOS  - ANO:"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__campo_ano = LiteralField(5, 81)
        campos: List[Field] = [
            FloatField(11, 3 + 11 * i, 3, format="E")
            for i in range(len(MESES_DF))
        ]
        self.__linha = Line(campos)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoSerieRuidosUHE):
            return False
        bloco: BlocoSerieRuidosUHE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "valor": tabela.flatten(),
                }
            )
            df["ano"] = ano
            df["mes"] = np.tile(
                np.arange(1, len(MESES_DF) + 1), tabela.shape[0]
            )
            df["serie"] = repete_vetor(list(range(1, i + 1)))
            return df[["ano", "serie", "mes", "valor"]]

        # Identifica o ano em questão
        linha = file.readline()
        ano = self.__campo_ano.read(linha)

        # Salta as linhas adicionais
        for _ in range(5):
            file.readline()

        # Variáveis auxiliares
        tabela = np.zeros((MAX_ANOS_HISTORICO, len(MESES_DF)))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            tabela[i, :] = self.__linha.read(linha)
            i += 1


class BlocoCorrelRuidosUHE(Block):
    """
    Bloco de informações do arquivo `parpvaz.dat`
    relativo às séries de autocorrelações
    dos ruídos por usina.
    """

    __slots__ = [
        "__linha",
    ]

    BEGIN_PATTERN = "CORRELOGRAMO DA SERIE DE RUIDOS"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_mes: List[Field] = [LiteralField(3, 1)]
        campo_ano: List[Field] = [LiteralField(4, 5)]
        correl: List[Field] = [
            FloatField(10, 9 + 10 * i, 5) for i in range(11)
        ]
        self.__linha = Line(campo_mes + campo_ano + correl)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelRuidosUHE):
            return False
        bloco: BlocoCorrelRuidosUHE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_vetor_anos(anos: List[str]) -> List[int]:
            # Descobre os anos pré e pós estudo
            numero_anos_pre = len([p for p in anos if p == "PRE"]) // 12
            numero_anos_pos = len([p for p in anos if p == "POS"]) // 12
            anos_estudo = [int(p) for p in anos if p not in ["PRE", "POS"]]
            # TODO - remover fallback de 1970 quando estiver o ano correto.
            # Descobre o primeiro ano de estudo
            primeiro_ano_estudo = (
                sorted(anos_estudo)[0] if len(anos_estudo) > 0 else 1970
            )
            # Descobre o último ano de estudo
            ultimo_ano_estudo = (
                sorted(anos_estudo)[-1] if len(anos_estudo) > 0 else 1970
            )
            indice_inicio_pos = anos.index("POS")
            # Substitui os anos pré e pós pelos valores específicos
            for a in range(numero_anos_pre):
                idx_i = 12 * a
                idx_f = idx_i + 12
                ano = primeiro_ano_estudo - (numero_anos_pre - a)
                anos[idx_i:idx_f] = [str(ano)] * 12
            for a in range(numero_anos_pos):
                idx_i = indice_inicio_pos + 12 * a
                idx_f = idx_i + 12
                ano = ultimo_ano_estudo + a + 1
                anos[idx_i:idx_f] = [str(ano)] * 12
            return [int(a) for a in anos]

        def converte_vetor_meses(meses: List[str]) -> List[int]:
            return [MESES_ABREV.index(m) + 1 for m in meses]

        def converte_tabela_em_df():
            anos_conv = converte_vetor_anos(anos)
            meses_conv = converte_vetor_meses(meses)
            datas = [
                date(year=a, month=m, day=1)
                for a, m in zip(anos_conv, meses_conv)
            ]
            df = pd.DataFrame(
                data={
                    "data": repete_vetor(datas, len(MESES_DF) - 1),
                    "lag": np.tile(np.arange(1, len(MESES_DF)), len(datas)),
                    "valor": tabela.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(4):
            file.readline()

        # Variáveis auxiliares
        meses: List[str] = []
        anos: List[str] = []
        tabela = np.zeros((MAX_ANOS_HISTORICO * len(MESES_DF), 11))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            meses.append(dados[0])
            anos.append(dados[1])
            tabela[i, :] = dados[2:]
            i += 1


class BlocoCorrelEspacialAnualMensalUHE(Block):
    """
    Bloco de informações do arquivo `parpvaz.dat`
    relativo à correlação espacial histórica anual e
    mensal para as UHEs.
    """

    __slots__ = [
        "__linha",
    ]

    BEGIN_PATTERN = "CORRELACAO ESPACIAL HISTORICA MENSAL/ANUAL ENTRE AS UHEs"
    END_PATTERN = ""

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        campo_uhe_1: List[Field] = [LiteralField(12, 1)]
        campo_uhe_2: List[Field] = [LiteralField(12, 15)]
        correl: List[Field] = [FloatField(6, 28 + 7 * i, 2) for i in range(13)]
        self.__linha = Line(campo_uhe_1 + campo_uhe_2 + correl)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoCorrelEspacialAnualMensalUHE):
            return False
        bloco: BlocoCorrelEspacialAnualMensalUHE = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(bloco.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(self, file: IO, *args, **kwargs):
        def converte_tabela_em_df():
            df = pd.DataFrame(tabela, columns=MESES_DF + ["anual"])
            df["uhe_1"] = uhes1
            df["uhe_2"] = uhes2
            return df[["uhe_1", "uhe_2"] + MESES_DF + ["anual"]]

        for _ in range(4):
            file.readline()

        # Variáveis auxiliares
        uhes1: List[str] = []
        uhes2: List[str] = []
        tabela = np.zeros((int(MAX_UHES * (MAX_UHES - 1) / 2), 13))
        i = 0
        while True:
            linha = file.readline()
            # Confere se acabou
            if len(linha) < 4:
                tabela = tabela[:i, :]
                self.data = converte_tabela_em_df()
                break
            # Senão, processa os dados
            dados = self.__linha.read(linha)
            uhes1.append(dados[0])
            uhes2.append(dados[1])
            tabela[i, :] = dados[2:]
            i += 1



from cfinterface.files.sectionfile import SectionFile
from inewave.newave.modelos.engnat import SecaoDadosEngnat
import pandas as pd  # type: ignore

from typing import TypeVar, Optional


class Engnat(SectionFile):
    """
    Armazena os dados de saída do NEWAVE referentes às séries históricas
    de energia por configuração, calculadas a partir das séries históricas
    de vazão.
    """

    T = TypeVar("T")

    SECTIONS = [SecaoDadosEngnat]
    STORAGE = "BINARY"

    @property
    def series(self) -> Optional[pd.DataFrame]:
        """
        Obtém a tabela com os dados das séries de energia
        afluente por REE e por estágio.

        - configuracao (`int`): configuração da série histórica
        - data (`datetime`): data para o valor histórico
        - ree (`int`): REE para o qual foi gerado
        - valor (`float`): energia em MWmes

        :return: A tabela com os dados das séries
        :rtype: pd.DataFrame | None
        """
        sections = [r for r in self.data.of_type(SecaoDadosEngnat)]
        if len(sections) > 0:
            return sections[0].data
        else:
            return None

    @series.setter
    def series(self, df: pd.DataFrame):
        sections = [r for r in self.data.of_type(SecaoDadosEngnat)]
        if len(sections) > 0:
            sections[0].data = df



from inewave.config import MAX_REES, MAX_VARIAVEIS_FOLGA_PENALIDADE

from cfinterface.components.section import Section
from cfinterface.components.line import Line
from cfinterface.components.literalfield import LiteralField
from cfinterface.components.integerfield import IntegerField
from cfinterface.components.floatfield import FloatField
from typing import List, IO
import pandas as pd  # type: ignore
import numpy as np  # type: ignore
from inewave._utils.formatacao import (
    repete_vetor,
)


class BlocoPenalidades(Section):
    """
    Bloco com informações das penalidades por violação de restrições
    e outras ações.
    """

    __slots__ = ["__linha", "__cabecalhos", "__numero_patamares_penalidade"]

    def __init__(self, previous=None, next=None, data=None) -> None:
        super().__init__(previous, next, data)
        self.__linha = Line(
            [
                LiteralField(6, 1),
                FloatField(8, 14, 2),
                FloatField(8, 24, 2),
                IntegerField(3, 36),
                IntegerField(2, 42),
                FloatField(8, 46, 2),
                FloatField(8, 56, 2),
            ]
        )
        self.__cabecalhos: List[str] = []

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, BlocoPenalidades):
            return False
        bloco: BlocoPenalidades = o
        if not all(
            [
                isinstance(self.data, pd.DataFrame),
                isinstance(o.data, pd.DataFrame),
            ]
        ):
            return False
        else:
            return self.data.equals(bloco.data)

    # Override
    def read(
        self, file: IO, numero_patamares_penalidade: int = 2, *args, **kwargs
    ):
        def converte_tabela_em_df():
            df = pd.DataFrame(
                data={
                    "variavel": repete_vetor(
                        chaves, self.__numero_patamares_penalidade
                    ),
                    "codigo_ree_submercado": repete_vetor(
                        codigos_ree_submercado,
                        self.__numero_patamares_penalidade,
                    ),
                    "patamar_penalidade": np.tile(
                        np.arange(1, self.__numero_patamares_penalidade + 1),
                        len(chaves),
                    ),
                    "patamar_carga": repete_vetor(
                        patamares_carga, self.__numero_patamares_penalidade
                    ),
                    "valor_R$_MWh": tabela_MWh.flatten(),
                    "valor_R$_hm3": tabela_hm3.flatten(),
                }
            )
            return df

        # Salta as linhas adicionais
        for _ in range(2):
            self.__cabecalhos.append(file.readline())

        self.__numero_patamares_penalidade = numero_patamares_penalidade

        i = 0
        chaves: List[str] = []
        codigos_ree_submercado: List[int] = []
        patamares_carga: List[int] = []
        tabela_MWh = np.zeros((MAX_REES * MAX_VARIAVEIS_FOLGA_PENALIDADE, 2))
        tabela_hm3 = np.zeros((MAX_REES * MAX_VARIAVEIS_FOLGA_PENALIDADE, 2))
        while True:
            linha = file.readline()
            # Confere se terminaram
            if len(linha) < 3:
                # Converte para df e salva na variável
                if i > 0:
                    tabela_MWh = tabela_MWh[:i, :]
                    tabela_hm3 = tabela_hm3[:i, :]
                    self.data = converte_tabela_em_df()
                break
            # Confere se é uma linha de subsistema ou tabela
            else:
                dados = self.__linha.read(linha)
                chaves.append(dados[0])
                codigos_ree_submercado.append(dados[3])
                patamares_carga.append(dados[4])
                tabela_MWh[i, :] = dados[1:3]
                tabela_hm3[i, :] = dados[5:7]
                i += 1

    # Override
    def write(self, file: IO, *args, **kwargs):
        for linha in self.__cabecalhos:
            file.write(linha)
        if not isinstance(self.data, pd.DataFrame):
            raise ValueError(
                "Dados do penalid.dat não foram lidos com sucesso"
            )

        df = self.data.copy()
        for _, linha_penalidade in (
            self.data[
                [
                    "variavel",
                    "codigo_ree_submercado",
                    "patamar_carga",
                ]
            ]
            .drop_duplicates()
            .iterrows()
        ):
            df_penalid = df.loc[df["variavel"] == linha_penalidade["variavel"]]
            if linha_penalidade["codigo_ree_submercado"] is not None:
                df_penalid = df_penalid.loc[
                    df_penalid["codigo_ree_submercado"]
                    == linha_penalidade["codigo_ree_submercado"]
                ]
            if linha_penalidade["patamar_carga"] is not None:
                df_penalid = df_penalid.loc[
                    df_penalid["patamar_carga"]
                    == linha_penalidade["patamar_carga"]
                ]
            file.write(
                self.__linha.write(
                    [
                        df_penalid["variavel"].iloc[0],
                    ]
                    + df_penalid.sort_values("patamar_penalidade")[
                        "valor_R$_MWh"
                    ].tolist()
                    + [
                        df_penalid["codigo_ree_submercado"].iloc[0],
                        df_penalid["patamar_carga"].iloc[0],
                    ]
                    + df_penalid.sort_values("patamar_penalidade")[
                        "valor_R$_hm3"
                    ].tolist()
                )
            )



MockBlocoPeriodoImpressao = [
    " INI FIM FC (FC = 1: IMPRIME TODOS CORTES, FC = 0: IMPRIME APENAS CORTES VALIDOS NA ULTIMA ITERACAO)\n",
    " XXX XXX X    \n",
    "   3  60 0\n",
]

MockBlocoOpcoesImpressao = [
    " OPCOES DE IMPRESSAO : 01 - CORTES FCF  02 - ESTADOS FCF  03 - RESTRICAO SAR\n",
    " XX XX XX (SE 99 CONSIDERA TODAS)\n",
    " 99\n",
]

MockNwlistcfdat = MockBlocoPeriodoImpressao + MockBlocoOpcoesImpressao



from inewave.nwlistop.modelos.blocos.usina import Usina
from inewave.nwlistop.modelos.arquivos.arquivousina import (
    ArquivoUsina,
)
from inewave.nwlistop.modelos.viol_pos_evap import ViolPosEvapAnos


class ViolPosEvap(ArquivoUsina):
    """
    Armazena os dados das saídas referentes ao desvios positivos da
    evaporação da usina.

    Esta classe lida com as informações de saída fornecidas pelo
    NWLISTOP e reproduzidas nos `viol_pos_evap00x.out`, onde x varia conforme
    a usina em questão.

    """

    BLOCKS = [
        Usina,
        ViolPosEvapAnos,
    ]
